import importlib
import os
import pkgutil
import sys
import typing as t
from collections import defaultdict
from typing import OrderedDict

import click
import yaml
from tabulate import tabulate


def from_yaml(filename: str):
    try:
        with open(filename, 'r') as f:
            return yaml.safe_load(f)
    except yaml.YAMLError as e:
        click.echo(f"Error reading YAML file: {e}", err=True)
        return {}


def dump_yaml(data, filename: str):
    try:
        with open(filename, 'w') as f:
            yaml.dump(data, f)
    except Exception as e:
        click.echo(f"Error writing YAML file: {e}", err=True)


# Dynamically find and import all command modules from the cli directory
def load_command_modules():
    package_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cli')

    for _, module_name, _ in pkgutil.iter_modules([package_dir]):
        if module_name not in ['base', '__main__']:  # Skip the base.py and __main__ file itself
            importlib.import_module(f'clarifai.cli.{module_name}')


def display_co_resources(
    response,
    custom_columns={
        'ID': lambda c: c.id,
        'USER_ID': lambda c: c.user_id,
        'DESCRIPTION': lambda c: c.description,
    },
):
    """Display compute orchestration resources listing results using rich."""

    formatter = TableFormatter(custom_columns)
    print(formatter.format(list(response), fmt="plain"))


class TableFormatter:
    def __init__(self, custom_columns: OrderedDict):
        """
        Initializes the TableFormatter with column headers and custom column mappings.

        :param headers: List of column headers for the table.
        """
        self.custom_columns = custom_columns

    def format(self, objects, fmt='plain'):
        """
        Formats a list of objects into a table with custom columns.

        :param objects: List of objects to format into a table.
        :return: A string representing the table.
        """
        # Prepare the rows by applying the custom column functions to each object
        rows = []
        for obj in objects:
            #   row = [self.custom_columns[header](obj) for header in self.headers]
            row = [f(obj) for f in self.custom_columns.values()]
            rows.append(row)

        # Create the table
        table = tabulate(rows, headers=self.custom_columns.keys(), tablefmt=fmt)
        return table


class AliasedGroup(click.Group):
    def __init__(
        self,
        name: t.Optional[str] = None,
        commands: t.Optional[
            t.Union[t.MutableMapping[str, click.Command], t.Sequence[click.Command]]
        ] = None,
        **attrs: t.Any,
    ) -> None:
        super().__init__(name, commands, **attrs)
        self.alias_map = {}
        self.command_to_aliases = defaultdict(list)

    def add_alias(self, cmd: click.Command, alias: str) -> None:
        self.alias_map[alias] = cmd
        if alias != cmd.name:
            self.command_to_aliases[cmd].append(alias)

    def command(
        self, aliases=None, *args, **kwargs
    ) -> t.Callable[[t.Callable[..., t.Any]], click.Command]:
        cmd_decorator = super().command(*args, **kwargs)
        if aliases is None:
            aliases = []

        def aliased_decorator(f):
            cmd = cmd_decorator(f)
            if cmd.name:
                self.add_alias(cmd, cmd.name)
            for alias in aliases:
                self.add_alias(cmd, alias)
            return cmd

        f = None
        if args and callable(args[0]):
            (f,) = args
        if f is not None:
            return aliased_decorator(f)
        return aliased_decorator

    def group(
        self, aliases=None, *args, **kwargs
    ) -> t.Callable[[t.Callable[..., t.Any]], click.Group]:
        cmd_decorator = super().group(*args, **kwargs)
        if aliases is None:
            aliases = []

        def aliased_decorator(f):
            cmd = cmd_decorator(f)
            if cmd.name:
                self.add_alias(cmd, cmd.name)
            for alias in aliases:
                self.add_alias(cmd, alias)
            return cmd

        f = None
        if args and callable(args[0]):
            (f,) = args
        if f is not None:
            return aliased_decorator(f)
        return aliased_decorator

    def get_command(self, ctx: click.Context, cmd_name: str) -> t.Optional[click.Command]:
        rv = click.Group.get_command(self, ctx, cmd_name)
        if rv is not None:
            return rv
        return self.alias_map.get(cmd_name)

    def format_commands(self, ctx, formatter):
        sub_commands = self.list_commands(ctx)

        rows = []
        for sub_command in sub_commands:
            cmd = self.get_command(ctx, sub_command)
            if cmd is None or getattr(cmd, 'hidden', False):
                continue
            if cmd in self.command_to_aliases:
                aliases = ', '.join(self.command_to_aliases[cmd])
                sub_command = f'{sub_command} ({aliases})'
            cmd_help = cmd.help
            rows.append((sub_command, cmd_help))

        if rows:
            with formatter.section("Commands"):
                formatter.write_dl(rows)


def validate_context(ctx):
    from clarifai.utils.logging import logger

    if ctx.obj == {}:
        logger.error("CLI config file missing. Run `clarifai login` to set up the CLI config.")
        sys.exit(1)



import os
import shutil

import click

from clarifai.cli.base import cli
from clarifai.utils.logging import logger


@cli.group(
    ['pipelinestep', 'ps'],
    context_settings={'max_content_width': shutil.get_terminal_size().columns - 10},
)
def pipeline_step():
    """Manage pipeline steps: upload, test, etc"""


@pipeline_step.command()
@click.argument("pipeline_step_path", type=click.Path(exists=True), required=False, default=".")
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help='Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile.',
)
def upload(pipeline_step_path, skip_dockerfile):
    """Upload a pipeline step to Clarifai.

    PIPELINE_STEP_PATH: Path to the pipeline step directory. If not specified, the current directory is used by default.
    """
    from clarifai.runners.pipeline_steps.pipeline_step_builder import upload_pipeline_step

    upload_pipeline_step(pipeline_step_path, skip_dockerfile)


@pipeline_step.command()
@click.argument(
    "pipeline_step_path",
    type=click.Path(),
    required=False,
    default=".",
)
def init(pipeline_step_path):
    """Initialize a new pipeline step directory structure.

    Creates the following structure in the specified directory:
    ├── 1/
    │   └── pipeline_step.py
    ├── requirements.txt
    └── config.yaml

    PIPELINE_STEP_PATH: Path where to create the pipeline step directory structure. If not specified, the current directory is used by default.
    """
    from clarifai.cli.templates.pipeline_step_templates import (
        get_config_template,
        get_pipeline_step_template,
        get_requirements_template,
    )

    # Resolve the absolute path
    pipeline_step_path = os.path.abspath(pipeline_step_path)

    # Create the pipeline step directory if it doesn't exist
    os.makedirs(pipeline_step_path, exist_ok=True)

    # Create the 1/ subdirectory
    pipeline_step_version_dir = os.path.join(pipeline_step_path, "1")
    os.makedirs(pipeline_step_version_dir, exist_ok=True)

    # Create pipeline_step.py
    pipeline_step_py_path = os.path.join(pipeline_step_version_dir, "pipeline_step.py")
    if os.path.exists(pipeline_step_py_path):
        logger.warning(f"File {pipeline_step_py_path} already exists, skipping...")
    else:
        pipeline_step_template = get_pipeline_step_template()
        with open(pipeline_step_py_path, 'w') as f:
            f.write(pipeline_step_template)
        logger.info(f"Created {pipeline_step_py_path}")

    # Create requirements.txt
    requirements_path = os.path.join(pipeline_step_path, "requirements.txt")
    if os.path.exists(requirements_path):
        logger.warning(f"File {requirements_path} already exists, skipping...")
    else:
        requirements_template = get_requirements_template()
        with open(requirements_path, 'w') as f:
            f.write(requirements_template)
        logger.info(f"Created {requirements_path}")

    # Create config.yaml
    config_path = os.path.join(pipeline_step_path, "config.yaml")
    if os.path.exists(config_path):
        logger.warning(f"File {config_path} already exists, skipping...")
    else:
        config_template = get_config_template()
        with open(config_path, 'w') as f:
            f.write(config_template)
        logger.info(f"Created {config_path}")

    logger.info(f"Pipeline step initialization complete in {pipeline_step_path}")
    logger.info("Next steps:")
    logger.info("1. Search for '# TODO: please fill in' comments in the generated files")
    logger.info("2. Update the pipeline step configuration in config.yaml")
    logger.info("3. Add your pipeline step dependencies to requirements.txt")
    logger.info("4. Implement your pipeline step logic in 1/pipeline_step.py")



import os
import re
import shutil
import subprocess
import urllib.parse
import uuid
from typing import Any, Dict, List

from clarifai_grpc.grpc.api.status import status_code_pb2

from clarifai.errors import UserError
from clarifai.utils.constants import HOME_PATH
from clarifai.utils.logging import logger

RETRYABLE_CODES = [
    status_code_pb2.MODEL_DEPLOYING,
    status_code_pb2.MODEL_LOADING,
    status_code_pb2.MODEL_BUSY_PLEASE_RETRY,
]

DEFAULT_CONFIG = HOME_PATH / '.config/clarifai/config'


def status_is_retryable(status_code: int) -> bool:
    """Check if a status code is retryable."""
    return status_code in RETRYABLE_CODES


class Chunker:
    """Split an input sequence into small chunks."""

    def __init__(self, seq: List, size: int) -> None:
        self.seq = seq
        self.size = size

    def chunk(self) -> List[List]:
        """Chunk input sequence."""
        return [self.seq[pos : pos + self.size] for pos in range(0, len(self.seq), self.size)]


class BackoffIterator:
    """Iterator that returns a sequence of backoff values."""

    def __init__(self, count=0):
        self.count = count

    def __iter__(self):
        return self

    def __next__(self):
        self.count += 1
        return 0.1 * (1.3**self.count)


def get_from_dict_or_env(key: str, env_key: str, **data) -> str:
    """Get a value from a dictionary or an environment variable."""
    if key in data and data[key]:
        return data[key]
    else:
        return get_from_env(key, env_key)


def get_from_env(key: str, env_key: str) -> str:
    """Get a value from a dictionary or an environment variable."""
    if env_key in os.environ and os.environ[env_key]:
        return os.environ[env_key]
    else:
        raise UserError(
            f"Did not find `{key}`, please add an environment variable"
            f" `{env_key}` which contains it, or pass"
            f"  `{key}` as a named parameter."
        )


def concept_relations_accumulation(
    relations_dict: Dict[str, Any], subject_concept: str, object_concept: str, predicate: str
) -> Dict[str, Any]:
    """Append the concept relation to relations dict based on its predicate.

    Args:
        relations_dict (dict): A dict of concept relations info.
    """
    if predicate == 'hyponym':
        if object_concept in relations_dict:
            relations_dict[object_concept].append(subject_concept)
        else:
            relations_dict[object_concept] = [subject_concept]
    elif predicate == 'hypernym':
        if subject_concept in relations_dict:
            relations_dict[subject_concept].append(object_concept)
        else:
            relations_dict[subject_concept] = [object_concept]
    else:
        relations_dict[object_concept] = []
        relations_dict[subject_concept] = []
    return relations_dict


def get_uuid(val: int) -> str:
    """Generates a UUID."""
    return uuid.uuid4().hex[:val]


def clean_input_id(input_id: str) -> str:
    """Clean input_id string into a valid input ID"""
    input_id = re.sub('[., /]+', '_', input_id)
    input_id = re.sub('[_]+', '_', input_id)
    input_id = re.sub('[-]+', '-', input_id)
    input_id = input_id.lower().strip('_-')
    input_id = re.sub('[^a-z0-9-_]+', '', input_id)
    return input_id


def format_github_repo_url(github_repo):
    """Format GitHub repository URL to a standard format."""
    if github_repo.startswith('http'):
        return github_repo
    elif '/' in github_repo and not github_repo.startswith('git@'):
        # Handle "user/repo" format
        return f"https://github.com/{github_repo}.git"
    else:
        return github_repo


def clone_github_repo(repo_url, target_dir, github_pat=None, branch=None):
    """Clone a GitHub repository with optional GitHub PAT authentication and branch specification."""
    # Handle local file paths - just copy instead of cloning
    if os.path.exists(repo_url):
        try:
            shutil.copytree(repo_url, target_dir, ignore=shutil.ignore_patterns('.git'))
            logger.info(f"Successfully copied local repository from {repo_url}")
            return True
        except Exception as e:
            logger.error(f"Failed to copy local repository: {e}")
            return False

    cmd = ["git", "clone"]

    # Add branch specification if provided
    if branch:
        cmd.extend(["-b", branch])

    # Handle authentication with GitHub PAT
    if github_pat:
        # Parse the URL and validate the hostname
        parsed_url = urllib.parse.urlparse(repo_url)
        if parsed_url.hostname == "github.com":
            # Insert GitHub PAT into the URL for authentication
            authenticated_url = f"https://{github_pat}@{parsed_url.netloc}{parsed_url.path}"
            cmd.append(authenticated_url)
        else:
            logger.error(f"Invalid repository URL: {repo_url}")
            return False
    else:
        cmd.append(repo_url)

    cmd.append(target_dir)

    try:
        result = subprocess.run(cmd, capture_output=True, text=True, check=True)
        if branch:
            logger.info(f"Successfully cloned repository from {repo_url} (branch: {branch})")
        else:
            logger.info(f"Successfully cloned repository from {repo_url}")
        return True
    except subprocess.CalledProcessError as e:
        logger.error(f"Failed to clone repository: {e.stderr}")
        return False



import json
import os
import shutil
import sys

import click
import yaml

from clarifai import __version__
from clarifai.utils.cli import AliasedGroup, TableFormatter, load_command_modules
from clarifai.utils.config import Config, Context
from clarifai.utils.constants import DEFAULT_BASE, DEFAULT_CONFIG, DEFAULT_UI
from clarifai.utils.logging import logger


# @click.group(cls=CustomMultiGroup)
@click.group(cls=AliasedGroup)
@click.version_option(version=__version__)
@click.option('--config', default=DEFAULT_CONFIG)
@click.pass_context
def cli(ctx, config):
    """Clarifai CLI"""
    ctx.ensure_object(dict)
    if os.path.exists(config):
        cfg = Config.from_yaml(filename=config)
        ctx.obj = cfg
    else:
        cfg = Config(
            filename=config,
            current_context='default',
            contexts={
                'default': Context(
                    'default',
                    CLARIFAI_PAT=os.environ.get('CLARIFAI_PAT', ''),
                    CLARIFAI_USER_ID=os.environ.get('CLARIFAI_USER_ID', ''),
                    CLARIFAI_API_BASE=DEFAULT_BASE,
                    CLARIFAI_UI=DEFAULT_UI,
                )
            },
        )
        try:
            cfg.to_yaml(config)
        except Exception:
            logger.warning(
                "Could not write configuration to disk. Could be a read only file system."
            )
        ctx.obj = cfg  # still have the default config even if couldn't write.


@cli.command()
@click.argument('shell', type=click.Choice(['bash', 'zsh']))
def shell_completion(shell):
    """Shell completion script"""
    os.system(f"_CLARIFAI_COMPLETE={shell}_source clarifai")


@cli.group(
    ['cfg'],
    cls=AliasedGroup,
    context_settings={'max_content_width': shutil.get_terminal_size().columns - 10},
)
def config():
    """Manage CLI configuration"""


@config.command(['e'])
@click.pass_context
def edit(ctx):
    """Edit the configuration file"""
    os.system(f'{os.environ.get("EDITOR", "vi")} {ctx.obj.filename}')


@config.command(['current'])
@click.option('-o', '--output-format', default='name', type=click.Choice(['name', 'json', 'yaml']))
@click.pass_context
def current_context(ctx, output_format):
    """Get the current context"""
    if output_format == 'name':
        print(ctx.obj.current_context)
    elif output_format == 'json':
        print(json.dumps(ctx.obj.contexts[ctx.obj.current_context].to_serializable_dict()))
    else:
        print(yaml.safe_dump(ctx.obj.contexts[ctx.obj.current_context].to_serializable_dict()))


@config.command(['list', 'ls'])
@click.option(
    '-o', '--output-format', default='wide', type=click.Choice(['wide', 'name', 'json', 'yaml'])
)
@click.pass_context
def get_contexts(ctx, output_format):
    """Get all contexts"""
    if output_format == 'wide':
        columns = {
            '': lambda c: '*' if c.name == ctx.obj.current_context else '',
            'NAME': lambda c: c.name,
            'USER_ID': lambda c: c.user_id,
            'API_BASE': lambda c: c.api_base,
            'PAT': lambda c: pat_display(c.pat),
        }
        additional_columns = set()
        for cont in ctx.obj.contexts.values():
            if 'env' in cont:
                for key in cont.to_column_names():
                    if key not in columns:
                        additional_columns.add(key)
        for key in sorted(additional_columns):
            columns[key] = lambda c, k=key: getattr(c, k) if hasattr(c, k) else ""

        formatter = TableFormatter(
            custom_columns=columns,
        )
        print(formatter.format(ctx.obj.contexts.values(), fmt="plain"))
    elif output_format == 'name':
        print('\n'.join(ctx.obj.contexts))
    elif output_format in ('json', 'yaml'):
        dicts = [v.__dict__ for c, v in ctx.obj.contexts.items()]
        for d in dicts:
            d.pop('pat')
        if output_format == 'json':
            print(json.dumps(dicts))
        elif output_format == 'yaml':
            print(yaml.safe_dump(dicts))


@config.command(['use'])
@click.argument('context-name', type=str)
@click.pass_context
def use_context(ctx, context_name):
    """Set the current context"""
    if context_name not in ctx.obj.contexts:
        raise click.UsageError('Context not found')
    ctx.obj.current_context = context_name
    ctx.obj.to_yaml()
    print(f'Set {context_name} as the current context')


@config.command(['cat'])
@click.option('-o', '--output-format', default='yaml', type=click.Choice(['yaml', 'json']))
@click.pass_obj
def dump(ctx_obj, output_format):
    """Dump the configuration to stdout"""
    if output_format == 'yaml':
        yaml.safe_dump(ctx_obj.to_dict(), sys.stdout)
    else:
        json.dump(ctx_obj.to_dict(), sys.stdout, indent=2)


@config.command(['cat'])
@click.pass_obj
def env(ctx_obj):
    """Print env vars. Use: eval "$(clarifai config env)" """
    ctx_obj.current.print_env_vars()


@cli.command()
@click.argument('api_url', default="https://api.clarifai.com")
@click.option('--user_id', required=False, help='User ID')
@click.pass_context
def login(ctx, api_url, user_id):
    """Login command to set PAT and other configurations."""

    name = input('context name (default: "default"): ')
    user_id = user_id if user_id is not None else input('user id: ')
    pat = input_or_default(
        'personal access token value (default: "ENVVAR" to get our of env var rather than config): ',
        'ENVVAR',
    )

    context = Context(
        name,
        CLARIFAI_API_BASE=api_url,
        CLARIFAI_USER_ID=user_id,
        CLARIFAI_PAT=pat,
    )

    if context.name == '':
        context.name = 'default'

    ctx.obj.contexts[context.name] = context
    ctx.obj.current_context = context.name

    ctx.obj.to_yaml()


@cli.group(cls=AliasedGroup)
def context():
    """Manage contexts"""


def pat_display(pat):
    return pat[:5] + "****"


def input_or_default(prompt, default):
    value = input(prompt)
    return value if value else default


@context.command()
@click.argument('name')
@click.option('--user-id', required=False, help='User ID')
@click.option('--base-url', required=False, help='Base URL')
@click.option('--pat', required=False, help='Personal access token')
@click.pass_context
def create(
    ctx,
    name,
    user_id=None,
    base_url=None,
    pat=None,
):
    """Create a new context"""
    if name in ctx.obj.contexts:
        print(f'{name} already exists')
        sys.exit(1)
    if not user_id:
        user_id = input('user id: ')
    if not base_url:
        base_url = input_or_default(
            'base url (default: https://api.clarifai.com): ', 'https://api.clarifai.com'
        )
    if not pat:
        pat = input_or_default(
            'personal access token value (default: "ENVVAR" to get our of env var rather than config): ',
            'ENVVAR',
        )

    context = Context(name, CLARIFAI_USER_ID=user_id, CLARIFAI_API_BASE=base_url, CLARIFAI_PAT=pat)
    ctx.obj.contexts[context.name] = context
    ctx.obj.to_yaml()


# write a click command to delete a context
@context.command(['rm'])
@click.argument('name')
@click.pass_context
def delete(ctx, name):
    """Delete a context"""
    if name not in ctx.obj.contexts:
        print(f'{name} is not a valid context')
        sys.exit(1)
    ctx.obj.contexts.pop(name)
    ctx.obj.to_yaml()
    print(f'{name} deleted')


@context.command()
@click.argument('name', type=str)
@click.pass_context
def use(ctx, name):
    """Set the current context"""
    if name not in ctx.obj.contexts:
        raise click.UsageError('Context not found')
    ctx.obj.current_context = name
    ctx.obj.to_yaml()
    print(f'Set {name} as the current context')


@cli.command()
@click.argument('script', type=str)
@click.option('--context', type=str, help='Context to use')
@click.pass_context
def run(ctx, script, context=None):
    """Execute a script with the current context's environment"""
    context = ctx.obj.current if not context else context
    cmd = f'CLARIFAI_USER_ID={context.user_id} CLARIFAI_API_BASE={context.base_url} CLARIFAI_PAT={context.pat} '
    cmd += ' '.join([f'{k}={v}' for k, v in context.env.items()])
    cmd += f' {script}'
    os.system(cmd)


# Import the CLI commands to register them
load_command_modules()


def main():
    cli()



import os
from enum import Enum
from typing import List, Tuple, Union

from clarifai.client.dataset import Dataset
from clarifai.client.model import Model

from .helpers import (
    MACRO_AVG,
    EvalType,
    _BaseEvalResultHandler,
    get_eval_type,
    make_handler_by_type,
)

try:
    import seaborn as sns
except ImportError:
    raise ImportError("Can not import seaborn. Please run `pip install seaborn` to install it")

try:
    import matplotlib.pyplot as plt
except ImportError:
    raise ImportError(
        "Can not import matplotlib. Please run `pip install matplotlib` to install it"
    )

try:
    import pandas as pd
except ImportError:
    raise ImportError("Can not import pandas. Please run `pip install pandas` to install it")

try:
    from loguru import logger
except ImportError:
    from ..logging import logger

__all__ = ['EvalResultCompare']


class CompareMode(Enum):
    MANY_MODELS_TO_ONE_DATA = 0
    ONE_MODEL_TO_MANY_DATA = 1


class EvalResultCompare:
    """Compare evaluation result of models against datasets.
    Note: The module will pick latest result on the datasets.
    and models must be same model type

    Args:
    ---
      models (Union[List[Model], List[str]]): List of Model or urls of models.
      datasets (Union[Dataset, List[Dataset], str, List[str]]): A single or List of Url or Dataset
      attempt_evaluate (bool): Evaluate when model is not evaluated with the datasets.
      auth_kwargs (dict): Additional auth keyword arguments to be passed to the Dataset and Model if using url(s)
    """

    def __init__(
        self,
        models: Union[List[Model], List[str]],
        datasets: Union[Dataset, List[Dataset], str, List[str]],
        attempt_evaluate: bool = False,
        eval_info: dict = None,
        auth_kwargs: dict = {},
    ):
        assert isinstance(models, list), ValueError("Expected list")

        if len(models) > 1:
            self.mode = CompareMode.MANY_MODELS_TO_ONE_DATA
            self.comparator = "Model"
            assert isinstance(datasets, Dataset) or (
                isinstance(datasets, list) and len(datasets) == 1
            ), (
                f"When comparing multiple models, must provide only one `datasets`. However got {datasets}"
            )
        else:
            self.mode = CompareMode.ONE_MODEL_TO_MANY_DATA
            self.comparator = "Dataset"

        # validate models
        if all(map(lambda x: isinstance(x, str), models)):
            models = [Model(each, **auth_kwargs) for each in models]
        elif not all(map(lambda x: isinstance(x, Model), models)):
            raise ValueError(
                f"Expected all models are list of string or list of Model, got {[type(each) for each in models]}"
            )
        # validate datasets
        if not isinstance(datasets, list):
            datasets = [
                datasets,
            ]
        if all(map(lambda x: isinstance(x, str), datasets)):
            datasets = [Dataset(each, **auth_kwargs) for each in datasets]
        elif not all(map(lambda x: isinstance(x, Dataset), datasets)):
            raise ValueError(
                f"Expected datasets must be str, list of string or Dataset, list of Dataset, got {[type(each) for each in datasets]}"
            )
        # Validate models vs datasets together
        self._eval_handlers: List[_BaseEvalResultHandler] = []
        self.model_type = None
        logger.info("Initializing models...")
        for model in models:
            model.load_info()
            model_type = model.model_info.model_type_id
            if not self.model_type:
                self.model_type = model_type
            else:
                assert self.model_type == model_type, (
                    f"Can not compare when model types are different, {self.model_type} != {model_type}"
                )
            m = make_handler_by_type(model_type)(model=model)
            logger.info(f"* {m.get_model_name(pretify=True)}")
            m.find_eval_id(
                datasets=datasets, attempt_evaluate=attempt_evaluate, eval_info=eval_info
            )
            self._eval_handlers.append(m)

    @property
    def eval_handlers(self):
        return self._eval_handlers

    def _loop_eval_handlers(self, func_name: str, **kwargs) -> Tuple[list, list]:
        """Run methods of `eval_handlers[...].model`

        Args:
          func_name (str): method name, see `_BaseEvalResultHandler` child classes
          kwargs: keyword arguments of the method

        Return:
          tuple:
            - list of outputs
            - list of comparator names

        """
        outs = []
        comparators = []
        logger.info(f'Running `{func_name}`')
        for _, each in enumerate(self.eval_handlers):
            for ds_index, _ in enumerate(each.eval_data):
                func = eval(f'each.{func_name}')
                out = func(index=ds_index, **kwargs)

                if self.mode == CompareMode.MANY_MODELS_TO_ONE_DATA:
                    name = each.get_model_name(pretify=True)
                else:
                    name = each.get_dataset_name_by_index(ds_index, pretify=True)
                if out is None:
                    logger.warning(
                        f'{self.comparator}:{name} does not have valid data for `{func_name}`'
                    )
                    continue
                comparators.append(name)
                outs.append(out)

        if self.mode == CompareMode.MANY_MODELS_TO_ONE_DATA:
            apps = set([comp.split('/')[0] for comp in comparators])
            if len(apps) == 1:
                comparators = ['/'.join(comp.split('/')[1:]) for comp in comparators]

        if not outs:
            logger.warning(f'Model type {self.model_type} does not support `{func_name}`')

        return outs, comparators

    def detailed_summary(
        self,
        confidence_threshold: float = 0.5,
        iou_threshold: float = 0.5,
        area: str = "all",
        bypass_const=False,
    ) -> Union[Tuple[pd.DataFrame, pd.DataFrame], None]:
        """
        Retrieve and compute popular metrics of model.

        Args:
          confidence_threshold (float): confidence threshold, applicable for classification and detection. Default is 0.5
          iou_threshold (float): iou threshold, support in range(0.5, 1., step=0.1) applicable for detection
          area (float): size of area, support {all, small, medium}, applicable for detection

        Return:
          None or tuple of dataframe: df summary per concept and total concepts

        """
        df = []
        total = []
        # loop over all eval_handlers/dataset and call its method
        outs, comparators = self._loop_eval_handlers(
            'detailed_summary',
            confidence_threshold=confidence_threshold,
            iou_threshold=iou_threshold,
            area=area,
            bypass_const=bypass_const,
        )
        for indx, out in enumerate(outs):
            _df, _total = out
            _df[self.comparator] = [comparators[indx] for _ in range(len(_df))]
            _total['Concept'].replace(
                to_replace=['Total'], value=f'{self.comparator}:{comparators[indx]}', inplace=True
            )
            _total.rename({'Concept': 'Total Concept'}, axis=1, inplace=True)
            df.append(_df)
            total.append(_total)

        if df:
            df = pd.concat(df, axis=0)
            total = pd.concat(total, axis=0)
            return df, total
        else:
            return None

    def confusion_matrix(
        self, show=True, save_path: str = None, cm_kwargs: dict = {}
    ) -> Union[pd.DataFrame, None]:
        """Return dataframe of confusion matrix
        Args:
            show (bool, optional): Show the chart. Defaults to True.
            save_path (str): path to save rendered chart.
            cm_kwargs (dict): keyword args of `eval_handler[...].model.cm_kwargs` method.
        Returns:
            None or pd.Dataframe, If models don't have confusion matrix, return None
        """
        outs, comparators = self._loop_eval_handlers("confusion_matrix", **cm_kwargs)
        all_dfs = []
        for _, (df, anchor) in enumerate(zip(outs, comparators)):
            df[self.comparator] = [anchor for _ in range(len(df))]
            all_dfs.append(df)

        if all_dfs:
            all_dfs = pd.concat(all_dfs, axis=0)
            if save_path or show:

                def _facet_heatmap(data, **kws):
                    data = data.dropna(axis=1)
                    data = data.drop(self.comparator, axis=1)
                    concepts = data.columns
                    colnames = pd.MultiIndex.from_arrays([concepts], names=['Predicted'])
                    data.columns = colnames
                    ax = sns.heatmap(
                        data, cmap='Blues', annot=True, annot_kws={"fontsize": 8}, **kws
                    )
                    ax.set_xticklabels(ax.get_xticklabels(), rotation=45, fontsize=6)
                    ax.set_yticklabels(ax.get_yticklabels(), fontsize=6, rotation=0)

                temp = all_dfs.copy()
                temp.columns = ["_".join(pair) for pair in temp.columns]
                with sns.plotting_context(font_scale=5.5):
                    g = sns.FacetGrid(
                        temp,
                        col=self.comparator,
                        col_wrap=3,
                        aspect=1,
                        height=3,
                        sharex=False,
                        sharey=False,
                    )
                    cbar_ax = g.figure.add_axes([0.92, 0.3, 0.02, 0.4])
                    g = g.map_dataframe(
                        _facet_heatmap, cbar_ax=cbar_ax, vmin=0, vmax=1, cbar=True, square=True
                    )
                    g.set_titles(col_template=str(self.comparator) + ':{col_name}', fontsize=5)
                    if show:
                        plt.show()
                    if save_path:
                        g.savefig(save_path)

        return all_dfs if isinstance(all_dfs, pd.DataFrame) else None

    @staticmethod
    def _set_default_kwargs(kwargs: dict, var_name: str, value):
        if var_name not in kwargs:
            kwargs.update({var_name: value})
        return kwargs

    @staticmethod
    def _setup_default_lineplot(df: pd.DataFrame, kwargs: dict):
        hue_order = df["concept"].unique().tolist()
        hue_order.remove(MACRO_AVG)
        hue_order.insert(0, MACRO_AVG)
        EvalResultCompare._set_default_kwargs(kwargs, "hue_order", hue_order)

        sizes = {}
        for each in hue_order:
            s = 1.5
            if each == MACRO_AVG:
                s = 4.0
            sizes.update({each: s})
        EvalResultCompare._set_default_kwargs(kwargs, "sizes", sizes)
        EvalResultCompare._set_default_kwargs(kwargs, "size", "concept")

        EvalResultCompare._set_default_kwargs(kwargs, "errorbar", None)
        EvalResultCompare._set_default_kwargs(kwargs, "height", 5)

        return kwargs

    def roc_curve_plot(
        self,
        show=True,
        save_path: str = None,
        roc_curve_kwargs: dict = {},
        relplot_kwargs: dict = {},
    ) -> Union[pd.DataFrame, None]:
        """Return dataframe of ROC curve
        Args:
            show (bool, optional): Show the chart. Defaults to True.
            save_path (str): path to save rendered chart.
            pr_curve_kwargs (dict): keyword args of `eval_handler[...].model.roc_curve` method.
            relplot_kwargs (dict): keyword args of `sns.relplot` except {data,x,y,hue,kind,col}. where x="fpr", y="tpr", hue="concept"
        Returns:
            None or pd.Dataframe, If models don't have ROC curve, return None
        """
        sns.set_palette("Paired")
        outs, comparator = self._loop_eval_handlers("roc_curve", **roc_curve_kwargs)
        all_dfs = []
        for _, (df, anchor) in enumerate(zip(outs, comparator)):
            df[self.comparator] = [anchor for _ in range(len(df))]
            all_dfs.append(df)

        if all_dfs:
            all_dfs = pd.concat(all_dfs, axis=0)
            if save_path or show:
                relplot_kwargs = self._setup_default_lineplot(all_dfs, relplot_kwargs)
                g = sns.relplot(
                    data=all_dfs,
                    x="fpr",
                    y="tpr",
                    hue='concept',
                    kind="line",
                    col=self.comparator,
                    **relplot_kwargs,
                )
                g.set_titles(col_template=str(self.comparator) + ':{col_name}', fontsize=5)
                if show:
                    plt.show()
                if save_path:
                    g.savefig(save_path)

        return all_dfs if isinstance(all_dfs, pd.DataFrame) else None

    def pr_plot(
        self,
        show=True,
        save_path: str = None,
        pr_curve_kwargs: dict = {},
        relplot_kwargs: dict = {},
    ) -> Union[pd.DataFrame, None]:
        """Return dataframe of PR curve
        Args:
            show (bool, optional): Show the chart. Defaults to True.
            save_path (str): path to save rendered chart.
            pr_curve_kwargs (dict): keyword args of `eval_handler[...].model.pr_curve` method.
            relplot_kwargs (dict): keyword args of `sns.relplot` except {data,x,y,hue,kind,col} where x="recall", y="precision", hue="concept"
        Returns:
            None or pd.Dataframe, If models don't have PR curve, return None
        """
        sns.set_palette("Paired")
        outs, comparator = self._loop_eval_handlers("pr_curve", **pr_curve_kwargs)
        all_dfs = []
        for _, (df, anchor) in enumerate(zip(outs, comparator)):
            df[self.comparator] = [anchor for _ in range(len(df))]
            all_dfs.append(df)

        if all_dfs:
            all_dfs = pd.concat(all_dfs, axis=0)
            if save_path or show:
                relplot_kwargs = self._setup_default_lineplot(all_dfs, relplot_kwargs)
                g = sns.relplot(
                    data=all_dfs,
                    x="recall",
                    y="precision",
                    hue='concept',
                    kind="line",
                    col=self.comparator,
                    **relplot_kwargs,
                )
                g.set_titles(col_template=str(self.comparator) + ':{col_name}', fontsize=5)
                if show:
                    plt.show()
                if save_path:
                    g.savefig(save_path)

        return all_dfs if isinstance(all_dfs, pd.DataFrame) else None

    def all(
        self,
        output_folder: str,
        confidence_threshold: float = 0.5,
        iou_threshold: float = 0.5,
        overwrite: bool = False,
        metric_kwargs: dict = {},
        pr_plot_kwargs: dict = {},
        roc_plot_kwargs: dict = {},
    ):
        """Run all comparison methods one by one:
        - detailed_summary
        - pr_curve (if applicable)
        - pr_plot
        - confusion_matrix (if applicable)
        And save to output_folder

        Args:
          output_folder (str): path to output
          confidence_threshold (float): confidence threshold, applicable for classification and detection. Default is 0.5.
          iou_threshold (float): iou threshold, support in range(0.5, 1., step=0.1) applicable for detection.
          overwrite (bool): overwrite result of output_folder.
          metric_kwargs (dict): keyword args for `eval_handler[...].model.{method}`, except for {confidence_threshold, iou_threshold}.
          roc_plot_kwargs (dict): for relplot_kwargs of `roc_curve_plot` method.
          pr_plot_kwargs (dict): for relplot_kwargs of `pr_plot` method.
        """
        eval_type = get_eval_type(self.model_type)
        area = metric_kwargs.pop("area", "all")
        bypass_const = metric_kwargs.pop("bypass_const", False)

        fname = f"conf-{confidence_threshold}"
        if eval_type == EvalType.DETECTION:
            fname = f"{fname}_iou-{iou_threshold}_area-{area}"

        def join_root(*args):
            return os.path.join(output_folder, *args)

        output_folder = join_root(fname)
        if os.path.exists(output_folder) and not overwrite:
            raise RuntimeError(
                f"{output_folder} exists. If you want to overwrite, set `overwrite=True`"
            )

        os.makedirs(output_folder, exist_ok=True)

        logger.info("Making summary tables...")
        dfs = self.detailed_summary(
            confidence_threshold=confidence_threshold,
            iou_threshold=iou_threshold,
            area=area,
            bypass_const=bypass_const,
        )
        if dfs is not None:
            concept_df, total_df = dfs
            concept_df.to_csv(join_root("concepts_summary.csv"))
            total_df.to_csv(join_root("total_summary.csv"))

        curve_metric_kwargs = dict(
            confidence_threshold=confidence_threshold, iou_threshold=iou_threshold
        )
        curve_metric_kwargs.update(metric_kwargs)

        self.roc_curve_plot(
            show=False,
            save_path=join_root("roc.jpg"),
            roc_curve_kwargs=curve_metric_kwargs,
            relplot_kwargs=roc_plot_kwargs,
        )

        self.pr_plot(
            show=False,
            save_path=join_root("pr.jpg"),
            pr_curve_kwargs=curve_metric_kwargs,
            relplot_kwargs=pr_plot_kwargs,
        )

        self.confusion_matrix(
            show=False, save_path=join_root("confusion_matrix.jpg"), cm_kwargs=curve_metric_kwargs
        )

        logger.info(f"Done. Your outputs are saved at {output_folder}")



import concurrent.futures

import fsspec

from clarifai.utils.logging import logger


def download_input(input, auth_helper=None):
    _download_input_data(input.data, auth_helper=auth_helper)
    if input.data.parts:
        for i in range(len(input.data.parts)):
            _download_input_data(input.data.parts[i].data, auth_helper=auth_helper)


def _download_input_data(input_data, auth_helper=None):
    """
    This function will download any urls that are not already bytes.

    Args:
        input_data: The input data containing URLs to download
        auth_helper: Optional ClarifaiAuthHelper instance for authentication
    """
    # Get auth headers if auth_helper is provided
    auth_kwargs = {}
    if auth_helper is not None:
        auth_kwargs = _get_auth_kwargs(auth_helper)

    if input_data.image.url and not input_data.image.base64:
        # Download the image
        with fsspec.open(input_data.image.url, 'rb', **auth_kwargs) as f:
            input_data.image.base64 = f.read()
    if input_data.video.url and not input_data.video.base64:
        # Download the video
        with fsspec.open(input_data.video.url, 'rb', **auth_kwargs) as f:
            input_data.video.base64 = f.read()
    if input_data.audio.url and not input_data.audio.base64:
        # Download the audio
        with fsspec.open(input_data.audio.url, 'rb', **auth_kwargs) as f:
            input_data.audio.base64 = f.read()
    if input_data.text.url and not input_data.text.raw:
        # Download the text
        with fsspec.open(input_data.text.url, 'r', **auth_kwargs) as f:
            input_data.text.raw = f.read()


def _get_auth_kwargs(auth_helper):
    """
    Convert ClarifaiAuthHelper metadata to fsspec-compatible kwargs.

    Args:
        auth_helper: ClarifaiAuthHelper instance

    Returns:
        dict: kwargs to pass to fsspec.open() for authentication
    """
    if auth_helper is None:
        return {}

    try:
        # Get authentication metadata from the helper
        metadata = auth_helper.metadata

        # Convert gRPC metadata tuples to HTTP headers dict
        headers = {}
        for key, value in metadata:
            # Skip non-auth headers
            if key in ('authorization', 'x-clarifai-session-token'):
                headers[key] = value

        # Return fsspec-compatible kwargs
        return {'client_kwargs': {'headers': headers}}
    except Exception as e:
        logger.warning(f"Failed to get authentication headers: {e}")
        return {}


def ensure_urls_downloaded(request, max_threads=128, auth_helper=None):
    """
    This function will download any urls that are not already bytes and parallelize with a thread pool.

    Args:
        request: The request containing inputs with URLs to download
        max_threads: Maximum number of threads to use for parallel downloads
        auth_helper: Optional ClarifaiAuthHelper instance for authentication
    """
    with concurrent.futures.ThreadPoolExecutor(max_workers=max_threads) as executor:
        futures = []
        for input in request.inputs:
            futures.append(executor.submit(download_input, input, auth_helper))
        for future in concurrent.futures.as_completed(futures):
            try:
                future.result()
            except Exception as e:
                logger.exception(f"Error downloading input: {e}")



"""Pipeline configuration validation utilities."""

import re
from typing import Any, Dict, List

import yaml


class PipelineConfigValidator:
    """Validator for pipeline configuration files."""

    # Regex patterns for templateRef validation
    TEMPLATE_REF_WITH_VERSION_PATTERN = re.compile(
        r'^users/([^/]+)/apps/([^/]+)/pipeline-steps/([^/]+)/versions/([^/]+)$'
    )
    TEMPLATE_REF_WITHOUT_VERSION_PATTERN = re.compile(
        r'^users/([^/]+)/apps/([^/]+)/pipeline-steps/([^/]+)$'
    )

    @classmethod
    def validate_config(cls, config: Dict[str, Any]) -> None:
        """Validate the pipeline configuration."""
        cls._validate_pipeline_section(config)
        cls._validate_orchestration_spec(config)

    @classmethod
    def _validate_pipeline_section(cls, config: Dict[str, Any]) -> None:
        """Validate the pipeline section of the config."""
        if "pipeline" not in config:
            raise ValueError("'pipeline' section not found in config.yaml")

        pipeline = config["pipeline"]
        required_fields = ["id", "user_id", "app_id"]

        for field in required_fields:
            if field not in pipeline:
                raise ValueError(f"'{field}' not found in pipeline section of config.yaml")
            if not pipeline[field]:
                raise ValueError(f"'{field}' cannot be empty in config.yaml")

        # Validate step_directories if present
        if "step_directories" in pipeline:
            if not isinstance(pipeline["step_directories"], list):
                raise ValueError("'step_directories' must be a list")

        # Validate orchestration_spec is present
        if "orchestration_spec" not in pipeline:
            raise ValueError("'orchestration_spec' not found in pipeline section")

    @classmethod
    def _validate_orchestration_spec(cls, config: Dict[str, Any]) -> None:
        """Validate the orchestration spec contains valid Argo workflow."""
        pipeline = config["pipeline"]
        orchestration_spec = pipeline["orchestration_spec"]

        if "argo_orchestration_spec" not in orchestration_spec:
            raise ValueError("'argo_orchestration_spec' not found in orchestration_spec")

        argo_spec_str = orchestration_spec["argo_orchestration_spec"]

        try:
            argo_spec = yaml.safe_load(argo_spec_str)
        except yaml.YAMLError as e:
            raise ValueError(f"Invalid YAML in argo_orchestration_spec: {e}")

        cls._validate_argo_workflow(argo_spec)

    @classmethod
    def _validate_argo_workflow(cls, argo_spec: Dict[str, Any]) -> None:
        """Validate the Argo workflow structure."""
        # Basic Argo workflow validation
        if not isinstance(argo_spec, dict):
            raise ValueError("argo_orchestration_spec must be a valid YAML object")

        required_fields = ["apiVersion", "kind", "spec"]
        for field in required_fields:
            if field not in argo_spec:
                raise ValueError(f"'{field}' not found in argo_orchestration_spec")

        if argo_spec["apiVersion"] != "argoproj.io/v1alpha1":
            raise ValueError("argo_orchestration_spec must have apiVersion 'argoproj.io/v1alpha1'")

        if argo_spec["kind"] != "Workflow":
            raise ValueError("argo_orchestration_spec must have kind 'Workflow'")

        # Validate templates and steps
        spec = argo_spec["spec"]
        if "templates" not in spec:
            raise ValueError("'templates' not found in argo_orchestration_spec.spec")

        cls._validate_argo_templates(spec["templates"])

    @classmethod
    def _validate_argo_templates(cls, templates: List[Dict[str, Any]]) -> None:
        """Validate Argo workflow templates."""
        for template in templates:
            if "steps" in template:
                for step_group in template["steps"]:
                    for step in step_group:
                        if "templateRef" in step:
                            template_ref = step["templateRef"]
                            cls._validate_template_ref(template_ref)

    @classmethod
    def _validate_template_ref(cls, template_ref: Dict[str, Any]) -> None:
        """Validate a templateRef in the Argo workflow."""
        if "name" not in template_ref or "template" not in template_ref:
            raise ValueError("templateRef must have both 'name' and 'template' fields")

        name = template_ref["name"]
        template = template_ref["template"]

        if name != template:
            raise ValueError(f"templateRef name '{name}' must match template '{template}'")

        # Check if it matches either pattern
        if not (
            cls.TEMPLATE_REF_WITH_VERSION_PATTERN.match(name)
            or cls.TEMPLATE_REF_WITHOUT_VERSION_PATTERN.match(name)
        ):
            raise ValueError(
                f"templateRef name '{name}' must match either pattern:\n"
                f"  - users/{{user_id}}/apps/{{app_id}}/pipeline-steps/{{step_id}}\n"
                f"  - users/{{user_id}}/apps/{{app_id}}/pipeline-steps/{{step_id}}/versions/{{version_id}}"
            )

    @classmethod
    def get_pipeline_steps_without_versions(cls, config: Dict[str, Any]) -> List[str]:
        """Get list of pipeline step names that don't have versions in templateRef."""
        pipeline = config["pipeline"]
        orchestration_spec = pipeline["orchestration_spec"]
        argo_spec_str = orchestration_spec["argo_orchestration_spec"]
        argo_spec = yaml.safe_load(argo_spec_str)

        steps_without_versions = []

        for template in argo_spec["spec"]["templates"]:
            if "steps" in template:
                for step_group in template["steps"]:
                    for step in step_group:
                        if "templateRef" in step:
                            template_ref = step["templateRef"]
                            name = template_ref["name"]

                            # Check if it's without version
                            if cls.TEMPLATE_REF_WITHOUT_VERSION_PATTERN.match(name):
                                # Extract step name
                                parts = name.split('/')
                                step_name = parts[-1]  # Last part is the step name
                                if step_name not in steps_without_versions:
                                    steps_without_versions.append(step_name)

        return steps_without_versions



import importlib
import inspect
import os
from typing import Dict, Tuple, Type, Union

from clarifai_grpc.grpc.api.service_pb2 import MultiDatasetVersionMetricsGroupResponse
from google.protobuf.json_format import MessageToDict

from clarifai.constants.dataset import TASK_TO_ANNOTATION_TYPE
from clarifai.datasets.upload.base import ClarifaiDataLoader
from clarifai.errors import UserError


def load_module_dataloader(module_dir: Union[str, os.PathLike], **kwargs) -> ClarifaiDataLoader:
    """Validate and import dataset module data generator.
    Args:
      `module_dir`: relative path to the module directory
      The directory must contain a `dataset.py` script and the data itself.
      kwargs: keyword arguments to be passed to the dataloader class
    Module Directory Structure:
    ---------------------------
        <folder_name>/
        ├──__init__.py
        ├──<Your local dir dataset>/
        └──dataset.py
    dataset.py must implement a class named following the convention,
    <dataset_name>DataLoader and this class must inherit from base ClarifaiDataLoader()
    """
    module_path = os.path.join(module_dir, "dataset.py")
    spec = importlib.util.spec_from_file_location("dataset", module_path)

    if not spec:
        raise ImportError(f"Module not found at {module_path}")

    # Load the module using the spec
    dataset = importlib.util.module_from_spec(spec)
    # Execute the module to make its contents available
    spec.loader.exec_module(dataset)

    # get main module class
    main_module_cls = None
    for name, obj in dataset.__dict__.items():
        if inspect.isclass(obj) and "DataLoader" in name:
            main_module_cls = obj
        else:
            continue

    return main_module_cls(**kwargs)


class DisplayUploadStatus:
    """Class to display dataset upload status."""

    def __init__(
        self,
        dataloader: ClarifaiDataLoader,
        dataset_metrics_response: Type[MultiDatasetVersionMetricsGroupResponse],
        dataset_info_dict: Dict[str, str],
        pre_upload_stats: Tuple[Dict[str, int], Dict[str, int]],
    ) -> None:
        """Initialize the class.
        Args:
          dataloader: ClarifaiDataLoader object
          dataset_metrics_response: The dataset version metrics response from the server.
          dataset_info_dict: The dataset info dictionary.
          pre_upload_stats: The pre upload stats for the dataset.
        """
        self.dataloader = dataloader
        self.dataset_metrics_response = dataset_metrics_response
        self.dataset_info_dict = dataset_info_dict
        self.pre_upload_stats = pre_upload_stats

        self.display()

    def display(self) -> None:
        """Display the upload status."""
        from rich.console import Console

        local_inputs_count, local_annotations_dict = self.get_dataloader_stats()
        uploaded_inputs_dict, uploaded_annotations_dict = self.get_dataset_version_stats(
            self.dataset_metrics_response
        )

        # Subtract the pre upload stats from the uploaded stats
        uploaded_inputs_dict = {
            key: int(uploaded_inputs_dict[key]) - int(self.pre_upload_stats[0].get(key, 0))
            for key in uploaded_inputs_dict
        }
        uploaded_annotations_dict = {
            key: uploaded_annotations_dict[key] - self.pre_upload_stats[1].get(key, 0)
            for key in uploaded_annotations_dict
        }

        self.local_annotations_count = sum(local_annotations_dict.values())
        self.uploaded_annotations_count = sum(uploaded_annotations_dict.values())

        local_dataset_dict = {
            "Inputs Count": str(local_inputs_count),
            "Annotations Count": str(local_annotations_dict),
        }
        uploaded_dataset_dict = {
            "Inputs Count": str(uploaded_inputs_dict["total"]),
            "Annotations Count": str(uploaded_annotations_dict),
        }

        panel_layout = self.get_display_layout(local_dataset_dict, uploaded_dataset_dict)

        console = Console()
        console.print(panel_layout)

    def get_dataloader_stats(self) -> Tuple[int, Dict[str, int]]:
        """Get the number of inputs and annotations in a dataloader.

        Returns:
          local_inputs_count (int): total number of inputs in the dataloader
          local_annotations_dict (Dict[str, int]): total number of annotations in the dataloader
        """
        from clarifai.constants.dataset import DATASET_UPLOAD_TASKS

        task = self.dataloader.task
        if task not in DATASET_UPLOAD_TASKS:
            raise UserError(
                "Invalid task, please use one of the following: {}".format(DATASET_UPLOAD_TASKS)
            )
        local_inputs_count = len(self.dataloader)
        local_annotations_dict = dict(concepts=0, bboxes=0, polygons=0)
        for i in range(local_inputs_count):
            key, attr = [(k, v) for k, v in TASK_TO_ANNOTATION_TYPE.get(task).items()][0]
            local_annotations_dict[key] += len(getattr(self.dataloader[i], attr))
        return local_inputs_count, local_annotations_dict

    @staticmethod
    def get_dataset_version_stats(
        dataset_metrics_response: Type[MultiDatasetVersionMetricsGroupResponse],
    ) -> Tuple[Dict[str, int], Dict[str, int]]:
        """Parse the response from the server for the dataset version metrics groups.
        Args:
          dataset_metrics_response: The dataset version metrics response from the server.

        Returns:
          uploaded_inputs_dict (Dict[str, int]): The input statistics for the dataset.
          uploaded_annotations_dict (Dict[str, int]): The annotation statistics for the dataset.
        """
        dataset_statistics = []
        uploaded_inputs_dict = {}
        uploaded_annotations_dict = dict(concepts=0, bboxes=0, polygons=0)
        dict_response = MessageToDict(dataset_metrics_response)

        for data in dict_response["datasetVersionMetricsGroups"]:
            if isinstance(data["value"], str):
                if ("type" in data) and (data["type"] == "CONCEPT_ID"):
                    data["metrics"].update({"Concept": data["value"]})
                    data["metrics"].pop("regionLocationMatrix", None)
                    dataset_statistics.append(data["metrics"])
                else:
                    uploaded_inputs_dict[data["value"]] = data["metrics"]["inputsCount"]

        for ds in dataset_statistics:
            uploaded_annotations_dict["bboxes"] += int(ds["boundingBoxesCount"])
            uploaded_annotations_dict["concepts"] += int(ds["positiveInputTagsCount"])
            uploaded_annotations_dict["polygons"] += int(ds["polygonsCount"])

        return uploaded_inputs_dict, uploaded_annotations_dict

    def _create_layout(self):
        from rich.layout import Layout
        from rich.progress import BarColumn, Progress, TextColumn

        # Create a Layout
        layout = Layout()

        # Add a new task to the progress bar
        progress = Progress(
            "{task.description}",
            BarColumn(bar_width=100),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        )

        # Split the layout into top and bottom rows
        layout.split(Layout(name="Progress"), Layout(name="Tables"))

        # Add the progress bar to the top layout
        layout["Progress"].update(progress)

        # Split the bottom layout into two columns
        layout["Tables"].split_row(
            Layout(name="Local Dataset"),
            Layout(name="Uploaded Dataset"),
        )

        # Create a new layout for the panels
        panel_layout = Layout(size=18)

        # Split the panel layout into top and bottom rows
        panel_layout.split(
            Layout(name="Progress Panel", size=9), Layout(name="Tables Panel", size=9)
        )

        return layout, panel_layout, progress

    def get_display_layout(
        self, local_dataset_dict: Dict[str, str], uploaded_dataset_dict: Dict[str, str]
    ):
        """Create a layout for the display.

        Args:
          local_dataset_dict (dict): The local dataset stats info dict.
          uploaded_dataset_dict (dict): The uploaded dataset stats info dict.

        Returns:
          panel_layout (Layout): The panel layout for the display.
        """
        from rich.console import Group
        from rich.panel import Panel

        from clarifai.utils.logging import table_from_dict

        local_dataset_table = table_from_dict(
            [local_dataset_dict],
            column_names=["Inputs Count", "Annotations Count"],
            title="[cyan]Local Dataset",
        )
        uploaded_dataset_table = table_from_dict(
            [uploaded_dataset_dict],
            column_names=["Inputs Count", "Annotations Count"],
            title="[cyan]Uploaded Dataset",
        )
        dataset_info_table = table_from_dict(
            [self.dataset_info_dict], column_names=["dataset_id", "user_id", "app_id"]
        )

        layout, panel_layout, progress = self._create_layout()

        # Add a new task to the progress bar
        progress.add_task(
            "[cyan]Inputs Progress:",
            completed=int(uploaded_dataset_dict["Inputs Count"]),
            total=int(local_dataset_dict["Inputs Count"]),
        )
        progress.add_task(
            "[cyan]Annotations Progress:",
            completed=self.uploaded_annotations_count,
            total=self.local_annotations_count,
        )

        # Add the tables to the respective layouts
        layout["Local Dataset"].update(local_dataset_table)
        layout["Uploaded Dataset"].update(uploaded_dataset_table)

        # Create a render group for the progress bar and the additional data
        progress_group = Group(progress, dataset_info_table)

        # Create a panel for the progress bar with a blue border and a suitable heading
        progress_panel = Panel(
            progress_group, title="[b] Dataset Upload Summary", border_style="blue"
        )

        # Create a panel for the tables comparison with a blue border and a suitable heading
        tables_panel = Panel(
            layout["Tables"],
            title="[b] Dataset Metrics Comparison",
            border_style="blue",
            expand=False,
        )

        # Add the panels to the respective layouts
        panel_layout["Progress Panel"].update(progress_panel)
        panel_layout["Tables Panel"].update(tables_panel)

        return panel_layout



import logging
import os
import time
import uuid

import pytest
from google.protobuf.struct_pb2 import Struct

from clarifai.client.user import User
from clarifai.datasets.upload.loaders.coco_detection import COCODetectionDataLoader
from clarifai.datasets.upload.loaders.imagenet_classification import ImageNetDataLoader
from clarifai.datasets.upload.utils import load_module_dataloader

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]
NOW = uuid.uuid4().hex[:10]
CREATE_APP_ID = f"ci_input_app_{NOW}"
CREATE_DATASET_ID = "ci_input_test_dataset"
CREATE_DATASET_NEW_ID = "ci_input_test_dataset_new"

# assets
IMAGE_URL = "https://samples.clarifai.com/metro-north.jpg"
VIDEO_URL = "https://samples.clarifai.com/beer.mp4"
TEXT_URL = "https://samples.clarifai.com/featured-models/Llama2_Conversational-agent.txt"
AUDIO_URL = "https://samples.clarifai.com/english_audio_sample.mp3"
IMAGE_FILE_PATH = os.path.dirname(__file__) + "/assets/red-truck.png"
VIDEO_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.mp4"
TEXT_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.txt"
AUDIO_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.mp3"
CSV_FILE_PATH = os.path.dirname(__file__) + "/assets/sample.csv"
FOLDER_PATH = os.path.dirname(__file__) + "/assets/test"
MODULE_DIR = os.path.dirname(__file__) + "/assets/voc"
COCO_DET_DIR = os.path.dirname(__file__) + "/assets/coco_detection"
IMAGENET_DIR = os.path.dirname(__file__) + "/assets/imagenet_classification"
TEXTS_FOLDER_PATH = os.path.dirname(__file__) + "/assets/sample_texts"

CLARIFAI_API_BASE = os.environ.get("CLARIFAI_API_BASE", "https://api.clarifai.com")


def client():
    return User(user_id=CREATE_APP_USER_ID, base_url=CLARIFAI_API_BASE)


@pytest.mark.requires_secrets
class Testdataupload:
    """Tests for data uploads.
    Uploads are tested for each of the following resources:
    - image
    - video
    - text
    - audio

    Tests for the following upload methods:
    - url
    - filepath
    - rawtext
    - CSV
    - Folder
    """

    @classmethod
    def setup_class(self):
        self.app = client().create_app(app_id=CREATE_APP_ID, base_workflow="Empty")
        self.input_object = self.app.inputs()
        self.dataset = self.app.create_dataset(dataset_id=CREATE_DATASET_ID)
        self.dataset_new = self.app.create_dataset(dataset_id=CREATE_DATASET_NEW_ID)

    def test_upload_image_url(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_url(input_id='input_1', image_url=IMAGE_URL)
            assert "SUCCESS" in caplog.text

    def test_upload_video_url(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_url(input_id='input_2', video_url=VIDEO_URL)
            assert "SUCCESS" in caplog.text

    def test_upload_text_url(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_url(input_id='input_3', text_url=TEXT_URL)
            assert "SUCCESS" in caplog.text

    def test_upload_audio_url(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_url(input_id='input_4', audio_url=AUDIO_URL)
            assert "SUCCESS" in caplog.text

    def test_upload_image_filepath(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_file(input_id='input_5', image_file=IMAGE_FILE_PATH)
            assert "SUCCESS" in caplog.text

    def test_upload_video_filepath(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_file(input_id='input_6', video_file=VIDEO_FILE_PATH)
            assert "SUCCESS" in caplog.text

    def test_upload_audio_filepath(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_file(input_id='input_7', audio_file=AUDIO_FILE_PATH)
            assert "SUCCESS" in caplog.text

    def test_upload_text_filepath(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_file(input_id='input_8', text_file=TEXT_FILE_PATH)
            assert "SUCCESS" in caplog.text

    def test_upload_rawtext(self, caplog):
        with caplog.at_level(logging.INFO):
            self.input_object.upload_text(input_id='input_9', raw_text='This is a test text')
            assert "SUCCESS" in caplog.text

    def test_upload_image_bytes(self, caplog):
        with open(IMAGE_FILE_PATH, 'rb') as image_file:
            image_bytes = image_file.read()
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_bytes(input_id='input_10', image_bytes=image_bytes)
            assert "SUCCESS" in caplog.text

    def test_upload_video_bytes(self, caplog):
        with open(VIDEO_FILE_PATH, 'rb') as video_file:
            video_bytes = video_file.read()
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_bytes(input_id='input_11', video_bytes=video_bytes)
            assert "SUCCESS" in caplog.text

    def test_upload_audio_bytes(self, caplog):
        with open(AUDIO_FILE_PATH, 'rb') as audio_file:
            audio_bytes = audio_file.read()
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_bytes(input_id='input_12', audio_bytes=audio_bytes)
            assert "SUCCESS" in caplog.text

    def test_upload_text_bytes(self, caplog):
        with open(TEXT_FILE_PATH, 'rb') as text_file:
            text_bytes = text_file.read()
        with caplog.at_level(logging.INFO):
            self.input_object.upload_from_bytes(input_id='input_13', text_bytes=text_bytes)
            assert "SUCCESS" in caplog.text

    def test_get_multimodal_input(self):
        input_object = self.input_object.get_multimodal_input(
            input_id='input_14', raw_text='This is a multimodal test text', image_url=IMAGE_URL
        )
        assert (
            input_object.id == 'input_14'
            and input_object.data.text.raw == 'This is a multimodal test text'
        )

    def test_get_text_inputs_from_folder(self):
        text_inputs = self.input_object.get_text_inputs_from_folder(TEXTS_FOLDER_PATH)
        assert len(text_inputs) == 3

    def test_get_input_from_app(self):
        input_object = self.input_object.get_input(input_id='input_1')
        assert input_object.id == 'input_1'

    def test_get_mask_proto(self):
        polygon_points = [[0.2, 0.2], [0.8, 0.2], [0.8, 0.8], [0.2, 0.8]]
        annotation = self.input_object.get_mask_proto(
            input_id="input_1",
            label="input_1_polygon_label",
            polygons=polygon_points,
            label_id="input_1_polygon_label",
            annot_id="input_1_polygon_annot",
        )
        assert annotation.id == "input_1_polygon_annot" and annotation.input_id == "input_1"

    def test_list_inputs(self):
        paginated_inputs = list(self.input_object.list_inputs(page_no=1, per_page=5))
        image_filterd_inputs = list(self.input_object.list_inputs(input_type='image'))
        downloaded_inputs = self.input_object.download_inputs(image_filterd_inputs)
        assert len(downloaded_inputs) == 3  # download inputs check
        assert len(paginated_inputs) == 5
        assert len(image_filterd_inputs) == 3  # 2 images uploaded in the above tests

    def test_patch_inputs(self):
        metadata = Struct()
        metadata.update({'test': 'SUCCESS'})
        new_input = self.input_object._get_proto(input_id='input_1', metadata=metadata)
        self.input_object.patch_inputs([new_input], action='merge')
        for input_item in list(self.input_object.list_inputs()):
            if input_item.id == 'input_1':
                assert input_item.data.metadata["test"] == "SUCCESS"
                break

    def test_patch_annotations(self, caplog):
        bbox_points = [0.2, 0.2, 0.8, 0.8]
        annotation = self.input_object.get_bbox_proto(
            input_id="input_1",
            label="input_1_label",
            bbox=bbox_points,
            label_id="input_1_label",
            annot_id="input_1_annot",
        )
        with caplog.at_level(logging.INFO):
            self.input_object.upload_annotations([annotation])
            assert "SUCCESS" in caplog.text  # upload annotations check

        bbox_points = [0.4, 0.4, 0.6, 0.6]
        annotation = self.input_object.get_bbox_proto(
            input_id="input_1",
            label="input_1_label",
            bbox=bbox_points,
            label_id="input_1_label",
            annot_id="input_1_annot",
        )
        self.input_object.patch_annotations([annotation], action='merge')
        test_annotation = list(self.input_object.list_annotations())[0]
        assert test_annotation.id == "input_1_annot"
        annot_bbox = test_annotation.data.regions[0].region_info.bounding_box
        assert (
            round(annot_bbox.left_col, 1) == 0.4
            and round(annot_bbox.top_row, 1) == 0.4
            and round(annot_bbox.right_col, 1) == 0.6
            and round(annot_bbox.bottom_row, 1) == 0.6
        )

    def test_patch_concepts(self):
        self.input_object.patch_concepts(
            concept_ids=["input_1_label"], labels=["SUCCESS"], values=[], action='overwrite'
        )
        concepts = list(self.app.list_concepts())
        for concept in concepts:
            if concept.id == "input_1_label":
                assert concepts[0].name == "SUCCESS"
                break

    def test_aggregate_inputs(self, caplog):
        uploaded_inputs = list(self.input_object.list_inputs())
        with caplog.at_level(logging.INFO):
            self.input_object.delete_inputs(uploaded_inputs)
            assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
        assert len(uploaded_inputs) == 13  # 9 inputs uploaded in the above tests

    def test_upload_csv(self, caplog):
        self.dataset.upload_from_csv(
            csv_path=CSV_FILE_PATH, input_type='text', csv_type='raw', labels=True
        )
        uploaded_inputs = list(self.input_object.list_inputs(dataset_id=CREATE_DATASET_ID))
        concepts = list(self.app.list_concepts())
        with caplog.at_level(logging.INFO):
            self.input_object.delete_inputs(uploaded_inputs)
            assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
        assert len(uploaded_inputs) == 5  # 5 inputs are uploaded from the CSV file
        assert len(concepts) == 3  # Test for list concepts

    def test_upload_folder(self, caplog):
        self.dataset.upload_from_folder(folder_path=FOLDER_PATH, input_type='image', labels=True)
        uploaded_inputs = list(self.input_object.list_inputs())
        with caplog.at_level(logging.INFO):
            self.input_object.delete_inputs(uploaded_inputs)
            assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
        assert (
            uploaded_inputs[0].data.concepts[0].name == 'test'
        )  # label of the first input in the folder
        assert len(uploaded_inputs) == 3  # 3 inputs are uploaded from the folder

    def test_upload_dataset(self, caplog):
        dataloader = load_module_dataloader(module_dir=MODULE_DIR, split="train")
        self.dataset.upload_dataset(dataloader, get_upload_status=True, log_warnings=True)
        uploaded_inputs = list(self.input_object.list_inputs())
        input_ids = [input_item.id for input_item in uploaded_inputs]
        annotations = list(self.input_object.list_annotations(batch_input=uploaded_inputs))
        with caplog.at_level(logging.INFO):
            self.input_object.delete_annotations(input_ids=input_ids)
            assert "Annotations Deleted" in caplog.text  # Testing delete annotations action
            self.input_object.delete_inputs(uploaded_inputs)
            assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
        assert len(uploaded_inputs) == 10  # 3 inputs are uploaded from the folder
        assert len(annotations) == 28  # Test for list annotations

    def test_upload_coco_dataset(self, caplog):
        dataloader = COCODetectionDataLoader(
            images_dir=os.path.join(COCO_DET_DIR, "images"),
            label_filepath=os.path.join(COCO_DET_DIR, "instances_default.json"),
        )
        self.dataset.upload_dataset(dataloader)
        uploaded_inputs = list(self.input_object.list_inputs())
        annotations = list(self.input_object.list_annotations(batch_input=uploaded_inputs))
        with caplog.at_level(logging.INFO):
            self.input_object.delete_inputs(uploaded_inputs)
            assert "Inputs Deleted" in caplog.text  # Testing delete inputs action
        assert len(uploaded_inputs) == 3  # 3 inputs are uploaded from the folder
        assert len(annotations) == 7  # Test for list annotations

    def test_upload_imagenet_dataset(self):
        dataloader = ImageNetDataLoader(data_dir=IMAGENET_DIR)
        self.dataset.upload_dataset(dataloader)
        uploaded_inputs = list(self.input_object.list_inputs())
        annotations = list(self.input_object.list_annotations(batch_input=uploaded_inputs))
        assert len(uploaded_inputs) == 10  # 10 inputs are uploaded from the folder
        assert len(annotations) == 10  # Test for list annotations

    def test_merge_datasets(self, caplog):
        with caplog.at_level(logging.INFO):
            self.dataset_new.merge_dataset(merge_dataset_id=CREATE_DATASET_ID)
            assert "SUCCESS" in caplog.text

    def test_export_dataset(self):
        dataset_demo_version = self.dataset.create_version()
        time.sleep(5)
        dataset_demo_version.export(save_path='tests/output_demo.zip')
        assert os.path.exists('tests/output_demo.zip') is True
        os.remove('tests/output_demo.zip')

    @classmethod
    def teardown_class(self):
        self.app.delete_dataset(dataset_id=CREATE_DATASET_ID)
        client().delete_app(app_id=CREATE_APP_ID)



import os
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed
from datetime import datetime
from multiprocessing import cpu_count
from typing import Dict, Generator, List, Optional, Tuple, Type, TypeVar, Union

import requests
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.resources_pb2 import Input
from clarifai_grpc.grpc.api.service_pb2 import MultiInputResponse
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from google.protobuf.json_format import MessageToDict
from requests.adapters import HTTPAdapter, Retry
from tabulate import tabulate
from tqdm import tqdm

from clarifai.client.base import BaseClient
from clarifai.client.input import Inputs
from clarifai.client.lister import Lister
from clarifai.constants.dataset import DATASET_UPLOAD_TASKS, MAX_RETRIES
from clarifai.datasets.export.inputs_annotations import (
    DatasetExportReader,
    InputAnnotationDownloader,
)
from clarifai.datasets.upload.base import ClarifaiDataLoader
from clarifai.datasets.upload.image import (
    VisualClassificationDataset,
    VisualDetectionDataset,
    VisualSegmentationDataset,
)
from clarifai.datasets.upload.multimodal import MultiModalDataset
from clarifai.datasets.upload.text import TextClassificationDataset
from clarifai.datasets.upload.utils import DisplayUploadStatus
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import add_file_handler, logger, process_log_files
from clarifai.utils.misc import BackoffIterator, Chunker

ClarifaiDatasetType = TypeVar(
    'ClarifaiDatasetType',
    VisualClassificationDataset,
    VisualDetectionDataset,
    VisualSegmentationDataset,
    TextClassificationDataset,
)
from clarifai.utils.constants import DEFAULT_BASE


class Dataset(Lister, BaseClient):
    """Dataset is a class that provides access to Clarifai API endpoints related to Dataset information."""

    def __init__(
        self,
        url: str = None,
        dataset_id: str = None,
        dataset_version_id: str = None,
        base_url: str = DEFAULT_BASE,
        pat: str = None,
        token: str = None,
        root_certificates_path: str = None,
        **kwargs,
    ):
        """Initializes a Dataset object.

        Args:
            url (str): The URL to initialize the dataset object.
            dataset_id (str): The Dataset ID within the App to interact with.
            dataset_version_id (str): The Dataset Version ID within the Dataset to interact with.
            base_url (str): Base API url. Default "https://api.clarifai.com"
            pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
            token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
            root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
            **kwargs: Additional keyword arguments to be passed to the Dataset.
        """
        if url and dataset_id:
            raise UserError("You can only specify one of url or dataset_id.")
        if url:
            user_id, app_id, _, dataset_id, dataset_version_id = (
                ClarifaiUrlHelper.split_clarifai_url(url)
            )
            kwargs = {'user_id': user_id, 'app_id': app_id}
        dataset_version = (
            {'id': dataset_version_id}
            if dataset_version_id
            else kwargs['version']
            if 'version' in kwargs
            else None
        )
        self.kwargs = {**kwargs, 'id': dataset_id, 'version': dataset_version}
        self.dataset_info = resources_pb2.Dataset(**self.kwargs)
        # Related to Dataset Upload
        self.num_workers: int = min(10, cpu_count())  # 15 req/sec rate limit
        self.annot_num_workers = 4
        self.max_retires = 10
        self.batch_size = 128  # limit max protos in a req
        self.task = None  # Upload dataset type
        self.input_object = Inputs(
            user_id=self.user_id,
            app_id=self.app_id,
            pat=pat,
            token=token,
            base_url=base_url,
            root_certificates_path=root_certificates_path,
        )
        self.logger = logger
        BaseClient.__init__(
            self,
            user_id=self.user_id,
            app_id=self.app_id,
            base=base_url,
            pat=pat,
            token=token,
            root_certificates_path=root_certificates_path,
        )
        Lister.__init__(self)

    def create_version(self, **kwargs) -> 'Dataset':
        """Creates a dataset version for the Dataset.

        Args:
            **kwargs: Additional keyword arguments to be passed to Dataset Version.
              - description (str): The description of the dataset version.
              - metadata (dict): The metadata of the dataset version.

        Returns:
            Dataset: A Dataset object for the specified dataset ID.

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
            >>> dataset_version = dataset.create_version(description='dataset_version_description')
        """
        request = service_pb2.PostDatasetVersionsRequest(
            user_app_id=self.user_app_id,
            dataset_id=self.id,
            dataset_versions=[resources_pb2.DatasetVersion(**kwargs)],
        )

        response = self._grpc_request(self.STUB.PostDatasetVersions, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nDataset Version created\n%s", response.status)
        kwargs.update(
            {
                'dataset_id': self.id,
                'version': response.dataset_versions[0],
            }
        )

        return Dataset.from_auth_helper(self.auth_helper, **kwargs)

    def delete_version(self, version_id: str) -> None:
        """Deletes a dataset version for the Dataset.

        Args:
            version_id (str): The version ID to delete.

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
            >>> dataset.delete_version(version_id='version_id')
        """
        request = service_pb2.DeleteDatasetVersionsRequest(
            user_app_id=self.user_app_id, dataset_id=self.id, dataset_version_ids=[version_id]
        )

        response = self._grpc_request(self.STUB.DeleteDatasetVersions, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nDataset Version Deleted\n%s", response.status)

    def list_versions(
        self, page_no: int = None, per_page: int = None
    ) -> Generator['Dataset', None, None]:
        """Lists all the versions for the dataset.

        Args:
            page_no (int): The page number to list.
            per_page (int): The number of items per page.

        Yields:
            Dataset: Dataset objects for the versions of the dataset.

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
            >>> all_dataset_versions = list(dataset.list_versions())

        Note:
            Defaults to 16 per page if page_no is specified and per_page is not specified.
            If both page_no and per_page are None, then lists all the resources.
        """
        request_data = dict(
            user_app_id=self.user_app_id,
            dataset_id=self.id,
        )
        all_dataset_versions_info = self.list_pages_generator(
            self.STUB.ListDatasetVersions,
            service_pb2.ListDatasetVersionsRequest,
            request_data,
            per_page=per_page,
            page_no=page_no,
        )

        for dataset_version_info in all_dataset_versions_info:
            dataset_version_info['id'] = dataset_version_info['dataset_version_id']
            del dataset_version_info['dataset_version_id']
            dataset_version_info.pop('metrics', None)
            dataset_version_info.pop('export_info', None)
            kwargs = {
                'dataset_id': self.id,
                'version': resources_pb2.DatasetVersion(**dataset_version_info),
            }
            yield Dataset.from_auth_helper(self.auth_helper, **kwargs)

    def list_inputs(
        self, page_no: int = None, per_page: int = None, input_type: str = None
    ) -> Generator[Input, None, None]:
        """Lists all the inputs for the dataset.

        Args:
            page_no (int): The page number to list.
            per_page (int): The number of items per page.
            input_type (str): The type of input to list. Options: 'image', 'video', 'audio', 'text'.

        Yields:
            Input: Input objects in the dataset.

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
            >>> all_dataset_inputs = list(dataset.list_inputs())
        """
        return self.input_object.list_inputs(
            dataset_id=self.id, page_no=page_no, per_page=per_page, input_type=input_type
        )

    def __iter__(self):
        return iter(DatasetExportReader(archive_url=self.archive_zip()))

    def _concurrent_annot_upload(
        self, annots: List[List[resources_pb2.Annotation]]
    ) -> Union[List[resources_pb2.Annotation], List[None]]:
        """Uploads annotations concurrently.

        Args:
          annots: annot protos

        Returns:
          retry_annot_upload: All failed annot protos during upload
        """
        annot_threads = []
        retry_annot_upload = []

        with ThreadPoolExecutor(
            max_workers=self.annot_num_workers
        ) as executor:  # limit annot workers
            annot_threads = [
                executor.submit(self.input_object.upload_annotations, inp_batch, False)
                for inp_batch in annots
            ]

            for job in as_completed(annot_threads):
                result = job.result()
                if result:
                    retry_annot_upload.extend(result)

        return retry_annot_upload

    def _delete_failed_inputs(
        self,
        batch_input_ids: List[int],
        dataset_obj: ClarifaiDatasetType,
        upload_response: MultiInputResponse = None,
        batch_no: Optional[int] = None,
    ) -> Tuple[List[int], List[int]]:
        """Delete failed input ids from clarifai platform dataset.

        Args:
          batch_input_ids: batch input ids
          dataset_obj: ClarifaiDataset object
          upload_response: upload response proto

        Returns:
          success_inputs: upload success input ids
          failed_inputs: upload failed input ids
        """
        success_status = status_pb2.Status(code=status_code_pb2.INPUT_DOWNLOAD_SUCCESS)
        input_ids = {dataset_obj.all_input_ids[id]: id for id in batch_input_ids}
        response = self._grpc_request(
            self.STUB.ListInputs,
            service_pb2.ListInputsRequest(
                ids=list(input_ids.keys()),
                per_page=len(input_ids),
                user_app_id=self.user_app_id,
                status=success_status,
            ),
        )
        response_dict = MessageToDict(response)
        success_inputs = response_dict.get('inputs', [])

        success_input_ids = [input.get('id') for input in success_inputs]
        failed_input_ids = list(set(input_ids) - set(success_input_ids.copy()))
        # check duplicate input ids
        duplicate_input_ids = [
            input.id
            for input in upload_response.inputs
            if input.status.details == 'Input has a duplicate ID.'
        ]  # handling duplicte ID failures.
        if duplicate_input_ids:
            success_input_ids = list(
                set(success_input_ids.copy()) - set(duplicate_input_ids.copy())
            )
            failed_input_ids = list(set(failed_input_ids) - set(duplicate_input_ids))
            duplicate_details = [
                [
                    input_ids[id],
                    id,
                    "Input has a duplicate ID.",
                    getattr(dataset_obj.data_generator[input_ids[id]], 'image_path', None),
                    getattr(dataset_obj.data_generator[input_ids[id]], 'labels', None),
                    getattr(dataset_obj.data_generator[input_ids[id]], 'metadata', None),
                ]
                for id in duplicate_input_ids
            ]
            duplicate_table = tabulate(
                duplicate_details,
                headers=["Index", "Input ID", "Status", "Image Path", "Labels", "Metadata"],
                tablefmt="grid",
            )
            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            self.logger.warning(
                f"{timestamp}\nFailed to upload {len(duplicate_input_ids)} inputs due to duplicate IDs in current batch {batch_no}:\n{duplicate_table}\n\n"
            )

        # delete failed inputs
        self._grpc_request(
            self.STUB.DeleteInputs,
            service_pb2.DeleteInputsRequest(user_app_id=self.user_app_id, ids=failed_input_ids),
        )
        return [input_ids[id] for id in success_input_ids], [
            input_ids[id] for id in failed_input_ids
        ]

    def _upload_inputs_annotations(
        self,
        batch_input_ids: List[int],
        dataset_obj: ClarifaiDatasetType,
        batch_no: Optional[int] = None,
        is_retry_duplicates: bool = False,
    ) -> Tuple[List[int], List[resources_pb2.Annotation], MultiInputResponse]:
        """Uploads batch of inputs and annotations concurrently to clarifai platform dataset.

        Args:
          batch_input_ids: batch input ids
          dataset_obj: ClarifaiDataset object

        Returns:
          failed_input_ids: failed input ids
          retry_annot_protos: failed annot protos
          response: upload response proto
        """
        input_protos, _ = dataset_obj.get_protos(batch_input_ids)
        if is_retry_duplicates:
            for inp in input_protos:
                inp.id = uuid.uuid4().hex

        input_job_id, _response = self.input_object.upload_inputs(
            inputs=input_protos, show_log=False
        )
        retry_annot_protos = []

        self.input_object._wait_for_inputs(input_job_id)
        success_input_ids, failed_input_ids = self._delete_failed_inputs(
            batch_input_ids, dataset_obj, _response, batch_no
        )

        if self.task in ["visual_detection", "visual_segmentation"] and success_input_ids:
            _, annotation_protos = dataset_obj.get_protos(success_input_ids)
            chunked_annotation_protos = Chunker(annotation_protos, self.batch_size).chunk()
            retry_annot_protos.extend(self._concurrent_annot_upload(chunked_annotation_protos))

        return failed_input_ids, retry_annot_protos, _response

    def _retry_uploads(
        self,
        failed_input_ids: List[int],
        retry_annot_protos: List[resources_pb2.Annotation],
        dataset_obj: ClarifaiDatasetType,
        batch_no: Optional[int],
    ) -> None:
        """Retry failed uploads.

        Args:
          failed_input_ids: failed input ids
          retry_annot_protos: failed annot protos
          dataset_obj: ClarifaiDataset object
        """

        for _retry in range(MAX_RETRIES):
            if not failed_input_ids and not retry_annot_protos:
                break
            if failed_input_ids:
                retry_input_ids = [dataset_obj.all_input_ids[id] for id in failed_input_ids]
                logger.warning(
                    f"Retrying upload for {len(failed_input_ids)} inputs in current batch: {retry_input_ids}\n"
                )
                failed_retrying_inputs, _, retry_response = self._upload_inputs_annotations(
                    failed_input_ids, dataset_obj, batch_no
                )
                failed_input_ids = failed_retrying_inputs
            if retry_annot_protos:
                chunked_annotation_protos = Chunker(retry_annot_protos, self.batch_size).chunk()
                _ = self._concurrent_annot_upload(chunked_annotation_protos)

        # Log failed inputs
        if failed_input_ids:
            failed_inputs_logs = []
            input_map = {input.id: input for input in retry_response.inputs}
            for index in failed_retrying_inputs:
                failed_id = dataset_obj.all_input_ids[index]
                input_details = input_map.get(failed_id)
                if input_details:
                    failed_input_details = [
                        index,
                        failed_id,
                        input_details.status.details,
                        getattr(dataset_obj.data_generator[index], 'image_path', None)
                        or getattr(dataset_obj.data_generator[index], 'text', None),
                        dataset_obj.data_generator[index].labels,
                        dataset_obj.data_generator[index].metadata,
                    ]
                    failed_inputs_logs.append(failed_input_details)

            failed_table = tabulate(
                failed_inputs_logs,
                headers=["Index", "Input ID", "Status", "Input", "Labels", "Metadata"],
                tablefmt="grid",
            )
            timestamp = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
            self.logger.warning(
                f"{timestamp}\nFailed to upload {len(failed_retrying_inputs)} inputs in current batch {batch_no} due to {retry_response}:\n{failed_table}\n\n"
            )

    def _data_upload(
        self,
        dataset_obj: ClarifaiDatasetType,
        is_log_retry: bool = False,
        log_retry_ids: List[int] = None,
        **kwargs,
    ) -> None:
        """Uploads inputs and annotations to clarifai platform dataset.

        Args:
          dataset_obj: ClarifaiDataset object,
          is_log_retry: True if the iteration is to retry uploads from logs.
          **kwargs: Additional keyword arguments for retry uploading functionality..

        Returns:
            None
        """
        if is_log_retry:
            input_ids = log_retry_ids
        else:
            input_ids = list(range(len(dataset_obj)))

        chunk_input_ids = Chunker(input_ids, self.batch_size).chunk()
        with ThreadPoolExecutor(max_workers=self.num_workers) as executor:
            with tqdm(total=len(chunk_input_ids), desc='Uploading Dataset') as progress:
                # Submit all jobs to the executor and store the returned futures
                futures = [
                    executor.submit(
                        self._upload_inputs_annotations,
                        batch_input_ids,
                        dataset_obj,
                        batch_no,
                        **kwargs,
                    )
                    for batch_no, batch_input_ids in enumerate(chunk_input_ids)
                ]

                for batch_no, job in enumerate(as_completed(futures)):
                    retry_input_ids, retry_annot_protos, _ = job.result()
                    self._retry_uploads(retry_input_ids, retry_annot_protos, dataset_obj, batch_no)
                    progress.update()

    def upload_dataset(
        self,
        dataloader: Type[ClarifaiDataLoader],
        batch_size: int = 32,
        get_upload_status: bool = False,
        log_warnings: bool = False,
        **kwargs,
    ) -> None:
        """Uploads a dataset to the app.

        Args:
          dataloader (Type[ClarifaiDataLoader]): ClarifaiDataLoader object
          batch_size (int): batch size for concurrent upload of inputs and annotations (max: 128)
          get_upload_status (bool): True if you want to get the upload status of the dataset
          log_warnings (bool): True if you want to save log warnings in a file
          kwargs: Additional keyword arguments for retry uploading functionality..
        """
        # set batch size and task
        self.batch_size = min(self.batch_size, batch_size)
        self.task = dataloader.task
        if self.task not in DATASET_UPLOAD_TASKS:
            raise UserError(
                "Task should be one of \
                      'text_classification', 'visual_classification', \
                      'visual_detection', 'visual_segmentation', 'visual_captioning', 'multimodal_dataset'"
            )

        if self.task == "text_classification":
            dataset_obj = TextClassificationDataset(dataloader, self.id)

        elif self.task == "visual_detection":
            dataset_obj = VisualDetectionDataset(dataloader, self.id)

        elif self.task == "visual_segmentation":
            dataset_obj = VisualSegmentationDataset(dataloader, self.id)

        elif self.task == "multimodal_dataset":
            dataset_obj = MultiModalDataset(dataloader, self.id)

        else:  # visual_classification & visual_captioning
            dataset_obj = VisualClassificationDataset(dataloader, self.id)

        if get_upload_status:
            pre_upload_stats = self.get_upload_status(pre_upload=True)

        # add file handler to log warnings
        if log_warnings:
            add_file_handler(
                self.logger, f"Dataset_Upload{str(int(datetime.now().timestamp()))}.log"
            )
        self._data_upload(dataset_obj, **kwargs)

        if get_upload_status:
            self.get_upload_status(dataloader=dataloader, pre_upload_stats=pre_upload_stats)

    def retry_upload_from_logs(
        self,
        log_file_path: str,
        dataloader: Type[ClarifaiDataLoader],
        retry_duplicates: bool = False,
        log_warnings: bool = False,
        **kwargs,
    ) -> None:
        """Retries failed uploads from the log file.

        Args:
            log_file_path (str): path to the log file
            dataloader (Type[ClarifaiDataLoader]): ClarifaiDataLoader object
            retry_duplicate (bool): True if you want to retry duplicate inputs
            kwargs: Additional keyword arguments for retry uploading functionality..
        """

        duplicate_input_ids, failed_input_ids = process_log_files(log_file_path)
        if log_warnings:
            add_file_handler(
                self.logger, f"Dataset_Upload{str(int(datetime.now().timestamp()))}.log"
            )

        if retry_duplicates and duplicate_input_ids:
            logger.warning(f"Retrying upload for {len(duplicate_input_ids)} duplicate inputs...\n")
            duplicate_inputs_indexes = [input["Index"] for input in duplicate_input_ids]
            self.upload_dataset(
                dataloader=dataloader,
                log_retry_ids=duplicate_inputs_indexes,
                is_retry_duplicates=True,
                is_log_retry=True,
                **kwargs,
            )

        if failed_input_ids:
            # failed_inputs= ([input["Input_ID"] for input in failed_input_ids])
            logger.warning(f"Retrying upload for {len(failed_input_ids)} failed inputs...\n")
            failed_input_indexes = [input["Index"] for input in failed_input_ids]
            self.upload_dataset(
                dataloader=dataloader,
                log_retry_ids=failed_input_indexes,
                is_log_retry=True,
                **kwargs,
            )

    def upload_from_csv(
        self,
        csv_path: str,
        input_type: str = 'text',
        csv_type: str = None,
        labels: bool = True,
        batch_size: int = 128,
    ) -> None:
        """Uploads dataset from a csv file.

        Args:
            csv_path (str): path to the csv file
            input_type (str): type of the dataset(text, image)
            csv_type (str): type of the csv file(raw, url, file_path)
            labels (bool): True if csv file has labels column
            batch_size (int): batch size for concurrent upload of inputs and annotations

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> dataset = Dataset(user_id = 'user_id', app_id = 'demo_app', dataset_id = 'demo_dataset')
            >>> dataset.upload_from_csv(csv_path='csv_path', input_type='text', csv_type='raw, labels=True)

        Note:
            CSV file supports 'inputid', 'input', 'concepts', 'metadata', 'geopoints' columns.
            All the data in the CSV should be in double quotes.
            metadata should be in single quotes format. Example: "{'key': 'value'}"
            geopoints should be in "long,lat" format.
        """
        if input_type not in ['image', 'text', 'video', 'audio']:
            raise UserError('Invalid input type, it should be image,text,audio or video')
        if csv_type not in ['raw', 'url', 'file_path']:
            raise UserError('Invalid csv type, it should be raw, url or file_path')
        assert csv_path.endswith('.csv'), 'csv_path should be a csv file'
        if csv_type == 'raw' and input_type != 'text':
            raise UserError('Only text input type is supported for raw csv type')
        batch_size = min(128, batch_size)
        input_protos = self.input_object.get_inputs_from_csv(
            csv_path=csv_path,
            input_type=input_type,
            csv_type=csv_type,
            dataset_id=self.id,
            labels=labels,
        )
        self.input_object._bulk_upload(inputs=input_protos, batch_size=batch_size)

    def upload_from_folder(
        self, folder_path: str, input_type: str, labels: bool = False, batch_size: int = 128
    ) -> None:
        """Upload dataset from folder.

        Args:
            folder_path (str): Path to the folder containing images.
            input_type (str): type of the dataset(text, image)
            labels (bool): True if folder name is the label for the inputs
            batch_size (int): batch size for concurrent upload of inputs and annotations

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> dataset = Dataset(user_id = 'user_id', app_id = 'demo_app', dataset_id = 'demo_dataset')
            >>> dataset.upload_from_folder(folder_path='folder_path', input_type='text', labels=True)

        Note: The filename is used as the input_id.
        """
        if input_type not in ['image', 'text']:
            raise UserError('Invalid input type it should be image or text')
        if input_type == 'image':
            input_protos = self.input_object.get_image_inputs_from_folder(
                folder_path=folder_path, dataset_id=self.id, labels=labels
            )
        if input_type == 'text':
            input_protos = self.input_object.get_text_inputs_from_folder(
                folder_path=folder_path, dataset_id=self.id, labels=labels
            )
        self.input_object._bulk_upload(inputs=input_protos, batch_size=batch_size)

    def get_upload_status(
        self,
        dataloader: Type[ClarifaiDataLoader] = None,
        delete_version: bool = False,
        timeout: int = 600,
        pre_upload_stats: Tuple[Dict[str, int], Dict[str, int]] = None,
        pre_upload: bool = False,
    ) -> Optional[Tuple[Dict[str, int], Dict[str, int]]]:
        """Creates a new dataset version and displays the upload status of the dataset.

        Args:
            dataloader (Type[ClarifaiDataLoader]): ClarifaiDataLoader object
            delete_version (bool): True if you want to delete the version after getting the upload status
            timeout (int): Timeout in seconds for getting the upload status. Default is 600 seconds.
            pre_upload_stats (Tuple[Dict[str, int], Dict[str, int]]): The pre upload stats for the dataset.
            pre_upload (bool): True if you want to get the pre upload stats for the dataset.

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
            >>> dataset.get_upload_status(dataloader)

        Note:
            This is a beta feature and is subject to change.
        """
        try:
            import rich  # noqa: F401
        except ImportError:
            raise UserError(
                "Rich library is not installed. Please install it using pip install rich>=13.4.2"
            )
        self.logger.info("Getting dataset upload status...")
        dataset_version_id = uuid.uuid4().hex
        _ = self.create_version(id=dataset_version_id, description="SDK Upload Status")

        request_data = dict(
            user_app_id=self.user_app_id,
            dataset_id=self.id,
            dataset_version_id=dataset_version_id,
        )

        start_time = time.time()
        backoff_iterator = BackoffIterator(10)
        while True:
            dataset_metrics_response = self._grpc_request(
                self.STUB.ListDatasetVersionMetricsGroups,
                service_pb2.ListDatasetVersionMetricsGroupsRequest(**request_data),
            )

            if dataset_metrics_response.status.code != status_code_pb2.SUCCESS:
                self.delete_version(dataset_version_id)
                raise Exception(
                    "Failed to get dataset metrics {}".format(dataset_metrics_response.status)
                )

            dict_response = MessageToDict(dataset_metrics_response)
            if len(dict_response.keys()) == 1 and time.time() - start_time < timeout:
                self.logger.info("Crunching the dataset metrics. Please wait...")
                time.sleep(next(backoff_iterator))
                continue
            else:
                if time.time() - start_time > timeout:
                    self.delete_version(dataset_version_id)
                    raise UserError(
                        "Dataset metrics are taking too long to process. Please try again later."
                    )
                break
        # get pre upload stats
        if pre_upload:
            return DisplayUploadStatus.get_dataset_version_stats(dataset_metrics_response)

        dataset_info_dict = dict(user_id=self.user_id, app_id=self.app_id, dataset_id=self.id)
        DisplayUploadStatus(
            dataloader, dataset_metrics_response, dataset_info_dict, pre_upload_stats
        )

        if delete_version:
            self.delete_version(dataset_version_id)

    def merge_dataset(self, merge_dataset_id: str) -> None:
        """Merges the another dataset into current dataset.

        Args:
            merge_dataset_id (str): The dataset ID of the dataset to merge.

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> dataset = Dataset(dataset_id='dataset_id', user_id='user_id', app_id='app_id')
            >>> dataset.merge_dataset(merge_dataset_id='merge_dataset_id')
        """
        dataset_filter = resources_pb2.Filter(
            input=resources_pb2.Input(dataset_ids=[merge_dataset_id])
        )
        query = resources_pb2.Search(query=resources_pb2.Query(filters=[dataset_filter]))
        request = service_pb2.PostDatasetInputsRequest(
            user_app_id=self.user_app_id, dataset_id=self.id, search=query
        )

        response = self._grpc_request(self.STUB.PostDatasetInputs, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        self.logger.info("\nDataset Merged\n%s", response.status)

    def archive_zip(self, wait: bool = True) -> str:
        """Exports the dataset to a zip file URL."""
        request = service_pb2.PutDatasetVersionExportsRequest(
            user_app_id=self.user_app_id,
            dataset_id=self.id,
            dataset_version_id=self.version.id,
            exports=[
                resources_pb2.DatasetVersionExport(
                    format=resources_pb2.DatasetVersionExportFormat.CLARIFAI_DATA_PROTOBUF
                )
            ],
        )

        response = self._grpc_request(self.STUB.PutDatasetVersionExports, request)
        if response.status.code != status_code_pb2.SUCCESS:
            raise Exception(response.status)
        if wait:
            while response.exports[0].status.code in (
                status_code_pb2.DATASET_VERSION_EXPORT_PENDING,
                status_code_pb2.DATASET_VERSION_EXPORT_IN_PROGRESS,
            ):
                time.sleep(1)
                response = self._grpc_request(self.STUB.PutDatasetVersionExports, request)
                if response.status.code != status_code_pb2.SUCCESS:
                    raise Exception(response.status)
        if response.exports[0].status.code != status_code_pb2.DATASET_VERSION_EXPORT_SUCCESS:
            raise Exception(response.exports[0].status)
        return response.exports[0].url

    def export(
        self,
        save_path: str,
        archive_url: str = None,
        local_archive_path: str = None,
        split: str = 'all',
        num_workers: int = 4,
    ) -> None:
        """Exports the Clarifai protobuf dataset to a local archive.

        Args:
            save_path (str): The path to save the archive to.
            archive_url (str): The URL to the Clarifai protobuf archive.
            local_archive_path (str): The path to the local Clarifai protobuf archive.
            split (str): Export dataset inputs in the directory format {split}/{input_type}. Default is all.
            num_workers (int): Number of workers to use for downloading the archive. Default is 4.

        Example:
            >>> from clarifai.client.dataset import Dataset
            >>> Dataset().export(save_path='output.zip')
        """
        if local_archive_path and not os.path.exists(local_archive_path):
            raise UserError(f"Archive {local_archive_path} does not exist.")
        if not archive_url and not local_archive_path:
            archive_url = self.archive_zip()
        # Create a session object and set auth header
        session = requests.Session()
        retries = Retry(total=5, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
        session.mount('https://', HTTPAdapter(max_retries=retries))
        session.headers.update({'Authorization': self.metadata[0][1]})
        with DatasetExportReader(
            session=session, archive_url=archive_url, local_archive_path=local_archive_path
        ) as reader:
            InputAnnotationDownloader(session, reader, num_workers).download_archive(
                save_path=save_path, split=split
            )

    def __getattr__(self, name):
        return getattr(self.dataset_info, name)

    def __str__(self):
        init_params = [param for param in self.kwargs.keys()]
        attribute_strings = [
            f"{param}={getattr(self.dataset_info, param)}"
            for param in init_params
            if hasattr(self.dataset_info, param)
        ]
        return f"Dataset Details: \n{', '.join(attribute_strings)}\n"



from schema import And, Optional, Regex, Schema


def get_schema() -> Schema:
    """Initialize the schema for rank and filter.

    This schema validates:

    - Rank and filter must be a list
    - Each item in the list must be a dict
    - The dict can contain these optional keys:
        - 'image_url': Valid URL string
        - 'text_raw': Non-empty string
        - 'metadata': Dict
        - 'image_bytes': Bytes
        - 'geo_point': Dict with 'longitude', 'latitude' and 'geo_limit' as float, float and int respectively
        - 'concepts': List where each item is a concept dict
    - Concept dict requires at least one of:
        - 'name': Non-empty string with dashes/underscores
        - 'id': Non-empty string
        - 'language': Non-empty string
        - 'value': 0 or 1 integer
    - 'input_types': List of 'image', 'video', 'text' or 'audio'
    - 'input_dataset_ids': List of strings
    - 'input_status_code': Integer

    Returns:
        Schema: The schema for rank and filter.
    """
    # Schema for a single concept
    concept_schema = Schema(
        {
            Optional('value'): And(int, lambda x: x in [0, 1]),
            Optional('id'): And(str, len),
            Optional('language'): And(str, len),
            # Non-empty strings with internal dashes and underscores.
            Optional('name'): And(str, len, Regex(r'^[0-9A-Za-z]+([-_][0-9A-Za-z]+)*$')),
        }
    )

    # Schema for a rank or filter item
    rank_filter_item_schema = Schema(
        {
            Optional('image_url'): And(str, Regex(r'^https?://')),
            Optional('text_raw'): And(str, len),
            Optional('metadata'): dict,
            Optional('image_bytes'): bytes,
            Optional('geo_point'): {'longitude': float, 'latitude': float, 'geo_limit': int},
            Optional("concepts"): And(
                list, lambda x: all(concept_schema.is_valid(item) and len(item) > 0 for item in x)
            ),
            ## input filters
            Optional('input_types'): And(
                list,
                lambda input_types: all(
                    input_type in ('image', 'video', 'text', 'audio') for input_type in input_types
                ),
            ),
            Optional('input_dataset_ids'): list,
            Optional('input_status_code'): int,
        }
    )

    # Schema for rank and filter args
    return Schema([rank_filter_item_schema])



from datetime import datetime
from typing import Any, Callable

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf import struct_pb2
from google.protobuf.timestamp_pb2 import Timestamp
from google.protobuf.wrappers_pb2 import BoolValue

from clarifai.client.auth import create_stub
from clarifai.client.auth.helper import ClarifaiAuthHelper
from clarifai.constants.base import COMPUTE_ORCHESTRATION_RESOURCES
from clarifai.errors import ApiError, UserError
from clarifai.utils.constants import CLARIFAI_PAT_ENV_VAR, CLARIFAI_SESSION_TOKEN_ENV_VAR
from clarifai.utils.misc import get_from_dict_or_env


class BaseClient:
    """BaseClient is the base class for all the classes interacting with Clarifai endpoints.

    Args:
        **kwargs: Additional keyword arguments to be passed to the ClarifaiAuthHelper.
            - user_id (str): A user ID for authentication.
            - app_id (str): An app ID for the application to interact with.
            - pat (str): A personal access token for authentication.
            - token (str): A session token for authentication. Accepts either a session token or a pat.
            - base (str): The base URL for the API endpoint. Defaults to 'https://api.clarifai.com'.
            - ui (str): The URL for the UI. Defaults to 'https://clarifai.com'.
            - root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.


    Attributes:
        auth_helper (ClarifaiAuthHelper): An instance of ClarifaiAuthHelper for authentication.
        STUB (Stub): The gRPC Stub object for API interaction.
        metadata (tuple): The gRPC metadata containing the personal access token.
        userDataObject (UserAppIDSet): The protobuf object representing user and app IDs.
        base (str): The base URL for the API endpoint.
    """

    def __init__(self, **kwargs):
        token, pat = "", ""
        try:
            pat = get_from_dict_or_env(key="pat", env_key=CLARIFAI_PAT_ENV_VAR, **kwargs)
        except UserError:
            token = get_from_dict_or_env(
                key="token", env_key=CLARIFAI_SESSION_TOKEN_ENV_VAR, **kwargs
            )
        finally:
            assert token or pat, Exception(
                "Need 'pat' or 'token' in args or use one of the CLARIFAI_PAT or CLARIFAI_SESSION_TOKEN env vars"
            )
        kwargs.update({'token': token, 'pat': pat})

        self.auth_helper = ClarifaiAuthHelper(**kwargs, validate=False)
        self.STUB = create_stub(self.auth_helper)
        self._async_stub = None
        self.metadata = self.auth_helper.metadata
        self.pat = self.auth_helper.pat
        self.token = self.auth_helper._token
        self.user_app_id = self.auth_helper.get_user_app_id_proto()
        self.base = self.auth_helper.base
        self.root_certificates_path = self.auth_helper._root_certificates_path

    @property
    def async_stub(self):
        """Returns the asynchronous gRPC stub for the API interaction.
        Lazy initialization of async stub"""
        if self._async_stub is None:
            self._async_stub = create_stub(self.auth_helper, is_async=True)
        return self._async_stub

    @classmethod
    def from_env(cls, validate: bool = False):
        auth = ClarifaiAuthHelper.from_env(validate=validate)
        return cls.from_auth_helper(auth)

    @classmethod
    def from_auth_helper(cls, auth: ClarifaiAuthHelper, **kwargs):
        default_kwargs = {
            "user_id": kwargs.get("user_id", None) or auth.user_id,
            "app_id": kwargs.get("app_id", None) or auth.app_id,
            "pat": kwargs.get("pat", None) or auth.pat,
            "token": kwargs.get("token", None) or auth._token,
            "root_certificates_path": kwargs.get("root_certificates_path", None)
            or auth._root_certificates_path,
        }
        _base = kwargs.get("base", None) or auth.base
        _clss = cls.__mro__[0]
        if _clss == BaseClient:
            kwargs = {
                **default_kwargs,
                "base": _base,  # Baseclient uses `base`
                "ui": kwargs.get("ui", None) or auth.ui,
            }
        else:
            # Remove user_id and app_id if a custom URL is provided
            if kwargs.get("url"):
                default_kwargs.pop("user_id", "")
                default_kwargs.pop("app_id", "")
            # Remove app_id if the class name is a compute orchestration resource
            if any(
                co_resource in _clss.__name__ for co_resource in COMPUTE_ORCHESTRATION_RESOURCES
            ):
                default_kwargs.pop("app_id", "")
            kwargs.update({**default_kwargs, "base_url": _base})

        return cls(**kwargs)

    def _grpc_request(self, method: Callable, argument: Any):
        """Makes a gRPC request to the API.

        Args:
            method (Callable): The gRPC method to call.
            argument (Any): The argument to pass to the gRPC method.

        Returns:
            res (Any): The result of the gRPC method call.
        """

        try:
            res = method(argument, metadata=self.auth_helper.metadata)
            # MessageToDict(res) TODO global debug logger
            return res
        except ApiError:
            raise Exception("ApiError")

    def convert_string_to_timestamp(self, date_str) -> Timestamp:
        """Converts a string to a Timestamp object.

        Args:
            date_str (str): The string to convert.

        Returns:
            Timestamp: The converted Timestamp object.
        """
        # Parse the string into a Python datetime object
        try:
            datetime_obj = datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%S.%fZ')
        except ValueError:
            try:
                datetime_obj = datetime.strptime(date_str, '%Y-%m-%dT%H:%M:%SZ')
            except ValueError:
                return Timestamp()

        # Convert the datetime object to a Timestamp object
        timestamp_obj = Timestamp()
        timestamp_obj.FromDatetime(datetime_obj)

        return timestamp_obj

    def process_response_keys(self, old_dict, listing_resource=None):
        """Converts keys in a response dictionary to resource proto format.

        Args:
            old_dict (dict): The dictionary to convert.
            listing_resource (str, optional): The resource type for which the keys are being processed.

        Returns:
            new_dict (dict): The dictionary with processed keys.
        """
        if listing_resource:
            old_dict[f'{listing_resource}_id'] = old_dict['id']
            old_dict.pop('id')

        def convert_recursive(item):
            if isinstance(item, dict):
                new_item = {}
                for key, value in item.items():
                    if key == 'default_value':
                        # Map infer param value to proto value
                        value_map = dict(number_value=None, string_value=None, bool_value=None)

                        def map_fn(v):
                            return (
                                'number_value'
                                if isinstance(v, float) or isinstance(v, int)
                                else 'string_value'
                                if isinstance(v, str)
                                else 'bool_value'
                                if isinstance(v, bool)
                                else None
                            )

                        value_map[map_fn(value)] = value
                        value = struct_pb2.Value(**value_map)
                    if key in ['created_at', 'modified_at', 'completed_at']:
                        value = self.convert_string_to_timestamp(value)
                    elif key in ['workflow_recommended', 'is_template']:
                        value = BoolValue(value=True)
                    elif key in ['fields_map', 'params']:
                        value_s = struct_pb2.Struct()
                        value_s.update(value)
                        value = value_s
                    elif key == 'metrics':
                        continue
                    elif key == 'size':
                        value = int(value)
                    elif key == 'image_info':
                        value = resources_pb2.ImageInfo(**value)
                    elif key == 'hosted_image_info':
                        continue
                    elif key in ['metadata', 'presets']:
                        if isinstance(value, dict) and value != {}:
                            value_s = struct_pb2.Struct()
                            value_s.update(value)
                            value = value_s
                        else:
                            continue
                    new_item[key] = convert_recursive(value)
                return new_item
            elif isinstance(item, list):
                return [convert_recursive(element) for element in item]
            else:
                return item

        new_dict = convert_recursive(old_dict)
        return new_dict



import glob
import importlib


class ClarifaiModulePageManager(object):
    def __init__(self):
        # List all the available pages.
        page_files = sorted(glob.glob("pages/*.py"))
        self.page_names = [f.replace("pages/", "").replace(".py", "") for f in page_files]

    def get_page_from_query_params(self, qp):
        """
        Args:
          qp: the streamlit query params st.experimental_get_query_params()
        """
        # Get the page from query params or default to 1 from the url.
        page = qp.get("page", [None])[0]
        if page is None:
            page = self.page_names[0]
        # Check that the page number coming in is within the range of pages in the folder.
        if page not in self.page_names:
            raise Exception(
                "Page '%s' is not valid, there is no pages/%s.py file for this page. Valid page names are: %s"
                % (page, page, str(self.page_names))
            )

        return page

    def get_page_names(self):
        return self.page_names

    def render_page(self, page):
        # Since the page re-renders every time the selectbox changes, we'll always have the latest page out
        # of the query params.
        module_str = "pages.%s" % page
        # check if the page exists
        importlib.util.find_spec(module_str)
        if page is None:
            raise Exception("Page %s is was not found." % page)

        current_page = importlib.import_module(module_str)
        current_page.display()



"""Base class for creating OpenAI-compatible API server."""

import json
from typing import Any, Dict, Iterator

from clarifai.runners.models.model_class import ModelClass


class OpenAIModelClass(ModelClass):
    """Base class for wrapping OpenAI-compatible servers as a model running in Clarifai.
    This handles all the transport between the API and the OpenAI-compatible server.

    To use this class, create a subclass and set the following class attributes:
    - client: The OpenAI-compatible client instance
    - model: The name of the model to use with the client

    Example:
        class MyOpenAIModel(OpenAIModelClass):
            client = OpenAI(api_key="your-key")
            model = "gpt-4"
    """

    # API Endpoints
    ENDPOINT_CHAT_COMPLETIONS = "/chat/completions"
    ENDPOINT_IMAGES_GENERATE = "/images/generations"
    ENDPOINT_EMBEDDINGS = "/embeddings"
    ENDPOINT_RESPONSES = "/responses"

    # Default endpoint
    DEFAULT_ENDPOINT = ENDPOINT_CHAT_COMPLETIONS

    # These should be overridden in subclasses
    client = None
    model = None

    def __init__(self) -> None:
        super().__init__()
        if self.client is None:
            raise NotImplementedError("Subclasses must set the 'client' class attribute")
        if self.model is None:
            try:
                self.model = self.client.models.list().data[0].id
            except Exception as e:
                raise NotImplementedError(
                    "Subclasses must set the 'model' class attribute or ensure the client can list models"
                ) from e

    def _create_completion_args(self, params: Dict[str, Any]) -> Dict[str, Any]:
        """Create the completion arguments dictionary from parameters.

        Args:
            params: Dictionary of parameters extracted from request

        Returns:
            Dict containing the completion arguments
        """
        completion_args = {**params}
        completion_args.update({"model": self.model})
        stream = completion_args.pop("stream", False)
        if stream:
            # Force to use usage
            stream_options = params.pop("stream_options", {})
            stream_options.update({"include_usage": True})
            completion_args["stream_options"] = stream_options
        completion_args["stream"] = stream

        return completion_args

    def _set_usage(self, resp):
        if resp.usage and resp.usage.prompt_tokens and resp.usage.completion_tokens:
            self.set_output_context(
                prompt_tokens=resp.usage.prompt_tokens,
                completion_tokens=resp.usage.completion_tokens,
            )

    def _handle_chat_completions(self, request_data: Dict[str, Any]):
        """Handle chat completion requests."""
        completion_args = self._create_completion_args(request_data)
        completion = self.client.chat.completions.create(**completion_args)
        self._set_usage(completion)
        return completion

    def _handle_images_generate(self, request_data: Dict[str, Any]):
        """Handle image generation requests."""
        image_args = {**request_data}
        image_args.update({"model": self.model})
        response = self.client.images.generate(**image_args)
        return response

    def _handle_embeddings(self, request_data: Dict[str, Any]):
        """Handle embedding requests."""
        embedding_args = {**request_data}
        embedding_args.update({"model": self.model})
        response = self.client.embeddings.create(**embedding_args)
        return response

    def _handle_responses(self, request_data: Dict[str, Any]):
        """Handle response requests."""
        response_args = {**request_data}
        response_args.update({"model": self.model})
        response = self.client.responses.create(**response_args)
        return response

    def _route_request(self, endpoint: str, request_data: Dict[str, Any]):
        """Route the request to appropriate handler based on endpoint."""
        handlers = {
            self.ENDPOINT_CHAT_COMPLETIONS: self._handle_chat_completions,
            self.ENDPOINT_IMAGES_GENERATE: self._handle_images_generate,
            self.ENDPOINT_EMBEDDINGS: self._handle_embeddings,
            self.ENDPOINT_RESPONSES: self._handle_responses,
        }

        handler = handlers.get(endpoint)
        if not handler:
            raise ValueError(f"Unsupported endpoint: {endpoint}")

        return handler(request_data)

    @ModelClass.method
    def openai_transport(self, msg: str) -> str:
        """Process an OpenAI-compatible request and send it to the appropriate OpenAI endpoint.

        Args:
            msg: JSON string containing the request parameters including 'openai_endpoint'

        Returns:
            JSON string containing the response or error
        """
        try:
            request_data = json.loads(msg)
            endpoint = request_data.pop("openai_endpoint", self.DEFAULT_ENDPOINT)
            response = self._route_request(endpoint, request_data)
            return json.dumps(response.model_dump())
        except Exception as e:
            return f"Error: {e}"

    @ModelClass.method
    def openai_stream_transport(self, msg: str) -> Iterator[str]:
        """Process an OpenAI-compatible request and return a streaming response iterator.
        This method is used when stream=True and returns an iterator of strings directly,
        without converting to a list or JSON serializing. Supports chat completions and responses endpoints.

        Args:
            msg: The request as a JSON string.

        Returns:
            Iterator[str]: An iterator yielding text chunks from the streaming response.
        """
        try:
            request_data = json.loads(msg)
            endpoint = request_data.pop("openai_endpoint", self.DEFAULT_ENDPOINT)
            if endpoint not in [self.ENDPOINT_CHAT_COMPLETIONS, self.ENDPOINT_RESPONSES]:
                raise ValueError("Streaming is only supported for chat completions and responses.")

            if endpoint == self.ENDPOINT_RESPONSES:
                # Handle responses endpoint
                stream_response = self._route_request(endpoint, request_data)
                for chunk in stream_response:
                    yield json.dumps(chunk.model_dump())
            else:
                completion_args = self._create_completion_args(request_data)
                stream_completion = self.client.chat.completions.create(**completion_args)
                for chunk in stream_completion:
                    self._set_usage(chunk)
                    yield json.dumps(chunk.model_dump())

        except Exception as e:
            yield f"Error: {e}"



import json
from typing import List

from clarifai_grpc.grpc.api import resources_pb2

from clarifai.runners.utils import data_utils
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.constants import MCP_TRANSPORT_NAME, OPENAI_TRANSPORT_NAME


def has_signature_method(
    name: str, method_signatures: List[resources_pb2.MethodSignature]
) -> bool:
    """
    Check if a method signature with the given name exists in the list of method signatures.

    :param name: The name of the method to check.
    :param method_signatures: List of MethodSignature objects to search in.
    :return: True if a method with the given name exists, False otherwise.
    """
    return any(
        method_signature.name == name for method_signature in method_signatures if method_signature
    )


def generate_client_script(
    method_signatures: List[resources_pb2.MethodSignature],
    user_id,
    app_id,
    model_id,
    base_url: str = None,
    deployment_id: str = None,
    compute_cluster_id: str = None,
    nodepool_id: str = None,
    use_ctx: bool = False,
) -> str:
    url_helper = ClarifaiUrlHelper()

    # Provide an mcp client config if there is a method named "mcp_transport"
    if has_signature_method(MCP_TRANSPORT_NAME, method_signatures):
        mcp_url = url_helper.mcp_api_url(
            user_id,
            app_id,
            model_id,
        )

        _CLIENT_TEMPLATE = """
import asyncio
import os
from fastmcp import Client
from fastmcp.client.transports import StreamableHttpTransport

transport = StreamableHttpTransport(url="%s",
                                    headers={"Authorization": "Bearer " + os.environ["CLARIFAI_PAT"]})

async def main():
  async with Client(transport) as client:
    tools = await client.list_tools()
    print(f"Available tools: {tools}")
    # TODO: update the dictionary of arguments passed to call_tool to make sense for your MCP.
    result = await client.call_tool(tools[0].name, {"a": 5, "b": 3})
    print(f"Result: {result[0].text}")

if __name__ == "__main__":
  asyncio.run(main())
"""
        return _CLIENT_TEMPLATE % mcp_url

    if has_signature_method(OPENAI_TRANSPORT_NAME, method_signatures):
        openai_api_base = url_helper.openai_api_url()
        model_ui_url = url_helper.clarifai_url(user_id, app_id, "models", model_id)
        _CLIENT_TEMPLATE = """
import os
from openai import OpenAI

client = OpenAI(
    base_url="%s",
    api_key=os.environ['CLARIFAI_PAT'],
)
response = client.chat.completions.create(
    model="%s",
    messages=[
        {"role": "system", "content": "Talk like a pirate."},
        {
            "role": "user",
            "content": "How do I check if a Python object is an instance of a class?",
        },
    ],
    temperature=0.7,
    stream=False, # stream=True also works, just iterator over the response
)
print(response)
"""
        return _CLIENT_TEMPLATE % (openai_api_base, model_ui_url)

    _CLIENT_TEMPLATE = """\
import os

from clarifai.client import Model
from clarifai.runners.utils import data_types
{model_section}
    """
    if deployment_id and (compute_cluster_id or nodepool_id):
        raise ValueError(
            "You can only specify one of deployment_id or compute_cluster_id and nodepool_id."
        )
    if compute_cluster_id and nodepool_id:
        deployment_id = None
    else:
        deployment_id = (
            "os.environ['CLARIFAI_DEPLOYMENT_ID']" if not deployment_id else repr(deployment_id)
        )

    deployment_line = (
        f'deployment_id = {deployment_id}, # Only needed for dedicated deployed models'
        if deployment_id
        else ""
    )
    compute_cluster_line = (
        f'compute_cluster_id = "{compute_cluster_id}",' if compute_cluster_id else ""
    )
    nodepool_line = (
        f'nodepool_id = "{nodepool_id}", # Only needed for dedicated nodepool'
        if nodepool_id
        else ""
    )

    base_url_str = ""
    if base_url is not None:
        base_url_str = f"base_url='{base_url}',"

    # Join all non-empty lines
    optional_lines = "\n    ".join(
        line
        for line in [deployment_line, compute_cluster_line, nodepool_line, base_url_str]
        if line
    )

    if use_ctx:
        model_section = """
model = Model.from_current_context()"""
    else:
        model_ui_url = url_helper.clarifai_url(user_id, app_id, "models", model_id)
        model_section = f"""
model = Model("{model_ui_url}",
    {optional_lines}
 )
"""

    # Generate client template
    client_template = _CLIENT_TEMPLATE.format(
        model_section=model_section,
    )

    # Generate method signatures
    method_signatures_str = []
    for method_signature in method_signatures:
        if method_signature is None:
            continue
        method_name = method_signature.name
        client_script_str = f'response = model.{method_name}('
        annotations = _get_annotations_source(method_signature)
        for idx, (param_name, (param_type, default_value, required)) in enumerate(
            annotations.items()
        ):
            if param_name == "return":
                continue
            if default_value is None and required:
                default_value = _set_default_value(param_type)
            if not default_value and idx == 0:
                default_value = _set_default_value(param_type)
            if param_type == "str" and default_value is not None:
                default_value = json.dumps(default_value)
            if default_value is not None:
                client_script_str += f"{param_name}={default_value}, "
        client_script_str = client_script_str.rstrip(", ") + ")"
        if method_signature.method_type == resources_pb2.RunnerMethodType.UNARY_UNARY:
            client_script_str += "\nprint(response)"
        elif method_signature.method_type == resources_pb2.RunnerMethodType.UNARY_STREAMING:
            client_script_str += "\nfor res in response:\n    print(res)"
        client_script_str += "\n"
        method_signatures_str.append(client_script_str)

    method_signatures_str = "\n".join(method_signatures_str)
    # Combine all parts
    script_lines = []
    script_lines.append("\n# Clarifai Model Client Script")
    if not use_ctx:
        script_lines.append(
            "# Set the environment variables `CLARIFAI_DEPLOYMENT_ID` and `CLARIFAI_PAT` to run this script."
        )
    script_lines.append("# Example usage:")
    script_lines.append(client_template)
    script_lines.append("# Example model prediction from different model methods: \n")
    script_lines.append(method_signatures_str)
    script_lines.append("")
    script = "\n".join(script_lines)
    return script


# get annotations source with default values
def _get_annotations_source(method_signature: resources_pb2.MethodSignature) -> dict:
    annotations = {}
    for input_field in method_signature.input_fields:
        param_name = input_field.name
        param_type = _get_base_type(input_field)
        if input_field.iterator:
            param_type = f"Iterator[{param_type}]"
        default_value = None
        if data_utils.Param.get_default(input_field):
            default_value = _parse_default_value(input_field)
        annotations[param_name] = (param_type, default_value, input_field.required)
    if not method_signature.output_fields:
        raise ValueError("MethodSignature must have at least one output field")
    for output_field in method_signature.output_fields:
        param_name = output_field.name
        param_type = _get_base_type(output_field)
        if output_field.iterator:
            param_type = f"Iterator[{param_type}]"
        annotations[param_name] = (param_type, None, output_field.required)
    return annotations


def _get_base_type(field: resources_pb2.ModelTypeField) -> str:
    data_type = field.type
    if data_type == resources_pb2.ModelTypeField.DataType.NAMED_FIELDS:
        annotations = {}
        for type_arg in field.type_args:
            field_name = type_arg.name
            field_type = _get_base_type(type_arg)
            annotations[field_name] = field_type
        class_name = f"NamedFields[{', '.join(f'{k}: {v}' for k, v in annotations.items())}]"
        return class_name
    elif data_type == resources_pb2.ModelTypeField.DataType.TUPLE:
        type_args_str = [_get_base_type(ta) for ta in field.type_args]
        return f"Tuple[{', '.join(type_args_str)}]"
    elif data_type == resources_pb2.ModelTypeField.DataType.LIST:
        if len(field.type_args) != 1:
            raise ValueError("List type must have exactly one type argument")
        element_type = _get_base_type(field.type_args[0])
        return f"List[{element_type}]"
    else:
        type_map = {
            resources_pb2.ModelTypeField.DataType.STR: "str",
            resources_pb2.ModelTypeField.DataType.BYTES: "bytes",
            resources_pb2.ModelTypeField.DataType.INT: "int",
            resources_pb2.ModelTypeField.DataType.FLOAT: "float",
            resources_pb2.ModelTypeField.DataType.BOOL: "bool",
            resources_pb2.ModelTypeField.DataType.NDARRAY: "np.ndarray",
            resources_pb2.ModelTypeField.DataType.JSON_DATA: "data_types.JSON",
            resources_pb2.ModelTypeField.DataType.TEXT: "data_types.Text",
            resources_pb2.ModelTypeField.DataType.IMAGE: "data_types.Image",
            resources_pb2.ModelTypeField.DataType.CONCEPT: "data_types.Concept",
            resources_pb2.ModelTypeField.DataType.REGION: "data_types.Region",
            resources_pb2.ModelTypeField.DataType.FRAME: "data_types.Frame",
            resources_pb2.ModelTypeField.DataType.AUDIO: "data_types.Audio",
            resources_pb2.ModelTypeField.DataType.VIDEO: "data_types.Video",
        }
        return type_map.get(data_type, "Any")


def _map_default_value(field_type):
    """
    Map the default value of a field type to a string representation.
    """
    default_value = None

    if field_type == "str":
        default_value = 'What is the future of AI?'
    elif field_type == "bytes":
        default_value = b""
    elif field_type == "int":
        default_value = 0
    elif field_type == "float":
        default_value = 0.0
    elif field_type == "bool":
        default_value = False
    elif field_type == "data_types.Image":
        default_value = 'data_types.Image.from_url("https://samples.clarifai.com/metro-north.jpg")'
    elif field_type == "data_types.Text":
        default_value = 'data_types.Text("What is the future of AI?")'
    elif field_type == "data_types.Audio":
        default_value = 'data_types.Audio.from_url("https://samples.clarifai.com/audio.mp3")'
    elif field_type == "data_types.Video":
        default_value = 'data_types.Video.from_url("https://samples.clarifai.com/video.mp4")'
    elif field_type == "data_types.Concept":
        default_value = 'data_types.Concept(id="concept_id", name="dog", value=0.95)'
    elif field_type == "data_types.Region":
        default_value = 'data_types.Region(box=[0.1, 0.1, 0.5, 0.5],)'
    elif field_type == "data_types.Frame":
        default_value = 'data_types.Frame.from_url("https://samples.clarifai.com/video.mp4", 0)'
    elif field_type == "data_types.NDArray":
        default_value = 'data_types.NDArray([1, 2, 3])'
    else:
        default_value = None
    return default_value


def _set_default_value(field_type):
    """
    Set the default value of a field if it is not set.
    """
    is_iterator = False
    if field_type.startswith("Iterator["):
        is_iterator = True
        field_type = field_type[9:-1]
    default_value = None
    default_value = _map_default_value(field_type)
    if field_type.startswith("List["):
        element_type = field_type[5:-1]
        element_type_default_value = _map_default_value(element_type)
        if element_type_default_value is not None:
            default_value = f"[{element_type_default_value}]"
    elif field_type.startswith("Tuple["):
        element_types = field_type[6:-1].split(", ")
        element_type_defaults = [_map_default_value(et) for et in element_types]
        default_value = f"({', '.join([str(et) for et in element_type_defaults])})"
    elif field_type.startswith("NamedFields["):
        element_types = field_type[12:-1].split(", ")
        element_type_defaults = [_map_default_value(et) for et in element_types]
        default_value = f"{{{', '.join([str(et) for et in element_type_defaults])}}}"

    if is_iterator:
        if field_type == "str":
            default_value = f"iter(['{default_value}'])"
        else:
            default_value = f"iter([{default_value}])"
    return default_value


def _parse_default_value(field: resources_pb2.ModelTypeField):
    if not field.default:
        return None
    default_str = field.default
    data_type = field.type

    try:
        if data_type == resources_pb2.ModelTypeField.DataType.INT:
            return str(int(default_str))
        elif data_type == resources_pb2.ModelTypeField.DataType.FLOAT:
            return str(float(default_str))
        elif data_type == resources_pb2.ModelTypeField.DataType.BOOL:
            return 'True' if default_str.lower() == 'true' else 'False'
        elif data_type == resources_pb2.ModelTypeField.DataType.STR:
            return default_str
        elif data_type == resources_pb2.ModelTypeField.DataType.BYTES:
            return f"b{repr(default_str.encode('utf-8'))}"
        elif data_type == resources_pb2.ModelTypeField.DataType.JSON_DATA:
            parsed = json.loads(default_str)
            return repr(parsed)
        elif data_type in (
            resources_pb2.ModelTypeField.DataType.LIST,
            resources_pb2.ModelTypeField.DataType.TUPLE,
            resources_pb2.ModelTypeField.DataType.NAMED_FIELDS,
        ):
            parsed = json.loads(default_str)
            return repr(parsed)
        else:
            return repr(default_str)
    except (ValueError, json.JSONDecodeError):
        return repr(default_str)
